# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
#
# Unit Tests for Elements of a Model
#
# TestScalarObj                Class for testing single objective
# TestArrayObj                Class for testing array of objective
#

import os
from os.path import abspath, dirname

currdir = dirname(abspath(__file__)) + os.sep

import pyomo.common.unittest as unittest
from pyomo.common.log import LoggingIntercept
from pyomo.environ import (
    ConcreteModel,
    AbstractModel,
    Objective,
    ObjectiveList,
    Var,
    Param,
    Set,
    RangeSet,
    value,
    maximize,
    minimize,
    simple_objective_rule,
    simple_objectivelist_rule,
)


class TestScalarObj(unittest.TestCase):
    def test_singleton_get_set(self):
        model = ConcreteModel()
        model.o = Objective(expr=1)
        self.assertEqual(len(model.o), 1)
        self.assertEqual(model.o.expr, 1)
        model.o.expr = 2
        self.assertEqual(model.o.expr, 2)
        model.o.expr += 2
        self.assertEqual(model.o.expr, 4)

    def test_singleton_get_set_value(self):
        model = ConcreteModel()
        model.o = Objective(expr=1)
        self.assertEqual(len(model.o), 1)
        self.assertEqual(model.o.expr, 1)
        model.o.expr = 2
        self.assertEqual(model.o.expr, 2)
        model.o.expr += 2
        self.assertEqual(model.o.expr, 4)

    def test_scalar_invalid_expr(self):
        m = ConcreteModel()
        m.x = Var()
        with self.assertRaisesRegex(
            ValueError,
            "Cannot assign InequalityExpression to 'obj': "
            "ScalarObjective components only allow numeric expression "
            "types.",
        ):
            m.obj = Objective(expr=m.x <= 0)

    def test_empty_singleton(self):
        a = Objective()
        a.construct()
        #
        # Even though we construct a ScalarObjective,
        # if it is not initialized that means it is "empty"
        # and we should encounter errors when trying to access the
        # ObjectiveData interface methods until we assign
        # something to the objective.
        #
        self.assertEqual(a._constructed, True)
        self.assertEqual(len(a), 0)
        with self.assertRaisesRegex(
            ValueError,
            "Evaluating the expression of ScalarObjective 'ScalarObjective' "
            "before the Objective has been assigned a sense or expression "
            r"\(there is currently no value to return\).",
        ):
            a()
        self.assertIsNone(a.expr)
        with self.assertRaisesRegex(
            ValueError,
            "Accessing the sense of ScalarObjective 'ScalarObjective' "
            "before the Objective has been assigned a sense or expression "
            r"\(there is currently no value to return\).",
        ):
            a.sense

        x = Var(initialize=1.0)
        x.construct()
        a.set_value(x + 1)
        self.assertEqual(len(a), 1)
        self.assertEqual(a(), 2)
        self.assertEqual(a.expr(), 2)
        self.assertEqual(a.sense, minimize)

    def test_unconstructed_singleton(self):
        a = Objective()
        self.assertEqual(a._constructed, False)
        self.assertEqual(len(a), 0)
        with self.assertRaisesRegex(
            RuntimeError,
            "Cannot access '__call__' on AbstractScalarObjective "
            "'AbstractScalarObjective' before it has been constructed",
        ):
            a()
        with self.assertRaisesRegex(
            RuntimeError,
            "Cannot access property 'expr' on AbstractScalarObjective "
            "'AbstractScalarObjective' before it has been constructed",
        ):
            a.expr
        with self.assertRaisesRegex(
            RuntimeError,
            "Cannot access property 'sense' on AbstractScalarObjective "
            "'AbstractScalarObjective' before it has been constructed",
        ):
            a.sense
        a.construct()
        a.set_sense(minimize)
        self.assertEqual(len(a), 1)
        self.assertEqual(a(), None)
        self.assertEqual(a.expr, None)
        self.assertEqual(a.sense, minimize)
        a.sense = maximize
        self.assertEqual(len(a), 1)
        self.assertEqual(a(), None)
        self.assertEqual(a.expr, None)
        self.assertEqual(a.sense, maximize)

    def test_numeric_expr(self):
        """Test expr option with a single numeric constant"""
        model = ConcreteModel()
        model.obj = Objective(expr=0.0)
        self.assertEqual(model.obj(), 0.0)
        self.assertEqual(value(model.obj), 0.0)
        self.assertEqual(value(model.obj._data[None]), 0.0)

    def test_mutable_param_expr(self):
        """Test expr option with a single mutable param"""
        model = ConcreteModel()
        model.p = Param(initialize=1.0, mutable=True)
        model.obj = Objective(expr=model.p)

        self.assertEqual(model.obj(), 1.0)
        self.assertEqual(value(model.obj), 1.0)
        self.assertEqual(value(model.obj._data[None]), 1.0)

    def test_immutable_param_expr(self):
        """Test expr option a single immutable param"""
        model = ConcreteModel()
        model.p = Param(initialize=1.0, mutable=False)
        model.obj = Objective(expr=model.p)

        self.assertEqual(model.obj(), 1.0)
        self.assertEqual(value(model.obj), 1.0)
        self.assertEqual(value(model.obj._data[None]), 1.0)

    def test_var_expr(self):
        """Test expr option with a single var"""
        model = ConcreteModel()
        model.x = Var(initialize=1.0)
        model.obj = Objective(expr=model.x)

        self.assertEqual(model.obj(), 1.0)
        self.assertEqual(value(model.obj), 1.0)
        self.assertEqual(value(model.obj._data[None]), 1.0)

    def test_expr1_option(self):
        """Test expr option"""
        model = ConcreteModel()
        model.B = RangeSet(1, 4)
        model.x = Var(model.B, initialize=2)
        ans = 0
        for i in model.B:
            ans = ans + model.x[i]
        model.obj = Objective(expr=ans)

        self.assertEqual(model.obj(), 8)
        self.assertEqual(value(model.obj), 8)
        self.assertEqual(value(model.obj._data[None]), 8)

    def test_expr2_option(self):
        """Test expr option"""
        model = ConcreteModel()
        model.x = Var(initialize=2)
        model.obj = Objective(expr=model.x)

        self.assertEqual(model.obj(), 2)
        self.assertEqual(value(model.obj), 2)
        self.assertEqual(value(model.obj._data[None]), 2)

    def test_rule_option(self):
        """Test rule option"""
        model = ConcreteModel()

        def f(model):
            ans = 0
            for i in [1, 2, 3, 4]:
                ans = ans + model.x[i]
            return ans

        model.x = Var(RangeSet(1, 4), initialize=2)
        model.obj = Objective(rule=f)

        self.assertEqual(model.obj(), 8)
        self.assertEqual(value(model.obj), 8)
        self.assertEqual(value(model.obj._data[None]), 8)

    def test_arguments(self):
        """Test that arguments notare of type ScalarSet"""
        model = ConcreteModel()

        def rule(model):
            return 1

        with self.assertRaisesRegex(
            TypeError,
            "Cannot apply a Set operator to a non-Set ConcreteModel component",
        ):
            model.obj = Objective(model, rule=rule)

    def test_sense_option(self):
        """Test sense option"""
        model = ConcreteModel()

        def rule(model):
            return 1.0

        model.obj = Objective(sense=maximize, rule=rule)

        self.assertEqual(model.obj.sense, maximize)
        self.assertEqual(model.obj.is_minimizing(), False)

    def test_dim(self):
        """Test dim method"""
        model = ConcreteModel()

        def rule(model):
            return 1

        model.obj = Objective(rule=rule)

        self.assertEqual(model.obj.dim(), 0)

    def test_keys(self):
        """Test keys method"""
        model = ConcreteModel()

        def rule(model):
            return 1

        model.obj = Objective(rule=rule)

        self.assertEqual(list(model.obj.keys()), [None])
        self.assertEqual(id(model.obj), id(model.obj[None]))

    def test_len(self):
        """Test len method"""
        model = AbstractModel()

        def rule(model):
            return 1.0

        model.obj = Objective(rule=rule)
        self.assertEqual(len(model.obj), 0)
        inst = model.create_instance()
        self.assertEqual(len(inst.obj), 1)

        model = AbstractModel()
        """Test rule option"""

        def f(model):
            ans = 0
            for i in model.x.keys():
                ans = ans + model.x[i]
            return ans

        model = AbstractModel()
        model.x = Var(RangeSet(1, 4), initialize=2)
        model.obj = Objective(rule=f)

        self.assertEqual(len(model.obj), 0)
        inst = model.create_instance()
        self.assertEqual(len(inst.obj), 1)

    def test_keys_empty(self):
        """Test keys method"""
        model = ConcreteModel()
        model.o = Objective()

        self.assertEqual(list(model.o.keys()), [])

    def test_len_empty(self):
        """Test len method"""
        model = ConcreteModel()
        model.o = Objective()
        self.assertEqual(len(model.o), 0)


class TestArrayObj(unittest.TestCase):
    def create_model(self):
        #
        # Create Model
        #
        model = ConcreteModel()
        model.A = Set(initialize=[1, 2])
        return model

    def test_objdata_get_set(self):
        model = ConcreteModel()
        model.o = Objective([1], rule=lambda m, i: 1)
        self.assertEqual(len(model.o), 1)
        self.assertEqual(model.o[1].expr, 1)
        model.o[1].expr = 2
        self.assertEqual(model.o[1].expr, 2)
        model.o[1].expr += 2
        self.assertEqual(model.o[1].expr, 4)

    def test_objdata_get_set_value(self):
        model = ConcreteModel()
        model.o = Objective([1], rule=lambda m, i: 1)
        self.assertEqual(len(model.o), 1)
        self.assertEqual(model.o[1].expr, 1)
        model.o[1].expr = 2
        self.assertEqual(model.o[1].expr, 2)
        model.o[1].expr += 2
        self.assertEqual(model.o[1].expr, 4)

    def test_objdata_get_set_sense(self):
        model = ConcreteModel()
        model.o = Objective([1], rule=lambda m, i: 1, sense=maximize)
        self.assertEqual(len(model.o), 1)
        self.assertEqual(model.o[1].expr, 1)
        self.assertEqual(model.o[1].sense, maximize)
        model.o[1].set_sense(minimize)
        self.assertEqual(model.o[1].sense, minimize)
        model.o[1].sense = maximize
        self.assertEqual(model.o[1].sense, maximize)

    def test_rule_option1(self):
        """Test rule option"""
        model = self.create_model()
        model.B = RangeSet(1, 4)

        def f(model, i):
            ans = 0
            for j in model.B:
                ans = ans + model.x[j]
            ans *= i
            return ans

        model.x = Var(model.B, initialize=2)
        model.obj = Objective(model.A, rule=f)

        self.assertEqual(model.obj[1](), 8)
        self.assertEqual(model.obj[2](), 16)
        self.assertEqual(value(model.obj[1]), 8)
        self.assertEqual(value(model.obj[2]), 16)

    def test_rule_option2(self):
        """Test rule option"""
        model = self.create_model()
        model.B = RangeSet(1, 4)

        def f(model, i):
            if i == 1:
                return Objective.Skip
            ans = 0
            for j in model.B:
                ans = ans + model.x[j]
            ans *= i
            return ans

        model.x = Var(model.B, initialize=2)
        model.obj = Objective(model.A, rule=f)

        self.assertEqual(model.obj[2](), 16)
        self.assertEqual(value(model.obj[2]), 16)

    def test_rule_option3(self):
        """Test rule option"""
        model = self.create_model()
        model.B = RangeSet(1, 4)

        @simple_objective_rule
        def f(model, i):
            if i == 1:
                return None
            ans = 0
            for j in model.B:
                ans = ans + model.x[j]
            ans *= i
            return ans

        model.x = Var(model.B, initialize=2)
        model.obj = Objective(model.A, rule=f)

        self.assertEqual(model.obj[2](), 16)
        self.assertEqual(value(model.obj[2]), 16)

    def test_rule_numeric_expr(self):
        """Test rule option with returns a single numeric constant for the expression"""
        model = self.create_model()

        def f(model, i):
            return 1.0

        model.obj = Objective(model.A, rule=f)

        self.assertEqual(model.obj[2](), 1.0)
        self.assertEqual(value(model.obj[2]), 1.0)

    def test_rule_immutable_param_expr(self):
        """Test rule option that returns a single immutable param for the expression"""
        model = self.create_model()

        def f(model, i):
            return model.p[i]

        model.p = Param(RangeSet(1, 4), initialize=1.0, mutable=False)
        model.x = Var()
        model.obj = Objective(model.A, rule=f)

        self.assertEqual(model.obj[2](), 1.0)
        self.assertEqual(value(model.obj[2]), 1.0)

    def test_rule_mutable_param_expr(self):
        """Test rule option that returns a single mutable param for the expression"""
        model = self.create_model()

        def f(model, i):
            return model.p[i]

        model.r = RangeSet(1, 4)
        model.p = Param(model.r, initialize=1.0, mutable=True)
        model.x = Var()
        model.obj = Objective(model.A, rule=f)

        self.assertEqual(model.obj[2](), 1.0)
        self.assertEqual(value(model.obj[2]), 1.0)

    def test_rule_var_expr(self):
        """Test rule option that returns a single var for the expression"""
        model = self.create_model()

        def f(model, i):
            return model.x[i]

        model.r = RangeSet(1, 4)
        model.x = Var(model.r, initialize=1.0)
        model.obj = Objective(model.A, rule=f)

        self.assertEqual(model.obj[2](), 1.0)
        self.assertEqual(value(model.obj[2]), 1.0)

    def test_sense_option(self):
        """Test sense option"""
        model = self.create_model()
        model.obj1 = Objective(model.A, rule=lambda m, i: 1.0, sense=maximize)
        model.obj2 = Objective(model.A, rule=lambda m, i: 1.0, sense=minimize)
        model.obj3 = Objective(model.A, rule=lambda m, i: 1.0)
        self.assertTrue(len(model.A) > 0)
        self.assertEqual(len(model.obj1), len(model.A))
        self.assertEqual(len(model.obj2), len(model.A))
        self.assertEqual(len(model.obj3), len(model.A))
        for i in model.A:
            self.assertEqual(model.obj1[i].sense, maximize)
            self.assertEqual(model.obj1[i].is_minimizing(), False)
            self.assertEqual(model.obj2[i].sense, minimize)
            self.assertEqual(model.obj2[i].is_minimizing(), True)
            self.assertEqual(model.obj3[i].sense, minimize)
            self.assertEqual(model.obj3[i].is_minimizing(), True)

    def test_dim(self):
        """Test dim method"""
        model = self.create_model()
        model.obj = Objective(model.A)

        self.assertEqual(model.obj.dim(), 1)

    def test_keys(self):
        """Test keys method"""
        model = self.create_model()

        def A_rule(model, i):
            return model.x

        model.x = Var()
        model.obj = Objective(model.A, rule=A_rule)

        self.assertEqual(len(list(model.obj.keys())), 2)

    def test_len(self):
        """Test len method"""
        model = self.create_model()
        model.obj = Objective(model.A)
        self.assertEqual(len(model.obj), 0)

        model = self.create_model()
        """Test rule option"""

        def f(model):
            ans = 0
            for i in model.x.keys():
                ans = ans + model.x[i]
            return ans

        model.x = Var(RangeSet(1, 4), initialize=2)
        model.obj = Objective(rule=f)

        self.assertEqual(len(model.obj), 1)


class Test2DArrayObj(unittest.TestCase):
    def create_model(self):
        model = ConcreteModel()
        model.A = Set(initialize=[1, 2])
        return model

    def test_rule_option1(self):
        """Test rule option"""
        model = self.create_model()
        model.B = RangeSet(1, 4)

        def f(model, i, k):
            ans = 0
            for j in model.B:
                ans = ans + model.x[j]
            ans *= i
            return ans

        model.x = Var(model.B, initialize=2)
        model.obj = Objective(model.A, model.A, rule=f)

        with self.assertRaisesRegex(
            TypeError, "'IndexedObjective' object is not callable"
        ):
            self.assertEqual(model.obj(), None)

        self.assertEqual(model.obj[1, 1](), 8)
        self.assertEqual(model.obj[2, 1](), 16)
        self.assertEqual(value(model.obj[1, 1]), 8)
        self.assertEqual(value(model.obj[2, 1]), 16)

    def test_sense_option(self):
        """Test sense option"""
        model = self.create_model()
        model.obj1 = Objective(
            model.A, model.A, rule=lambda m, i, j: 1.0, sense=maximize
        )
        model.obj2 = Objective(
            model.A, model.A, rule=lambda m, i, j: 1.0, sense=minimize
        )
        model.obj3 = Objective(model.A, model.A, rule=lambda m, i, j: 1.0)
        self.assertTrue(len(model.A) > 0)
        self.assertEqual(len(model.obj1), len(model.A) * len(model.A))
        self.assertEqual(len(model.obj2), len(model.A) * len(model.A))
        self.assertEqual(len(model.obj3), len(model.A) * len(model.A))
        for i in model.A:
            for j in model.A:
                self.assertEqual(model.obj1[i, j].sense, maximize)
                self.assertEqual(model.obj1[i, j].is_minimizing(), False)
                self.assertEqual(model.obj2[i, j].sense, minimize)
                self.assertEqual(model.obj2[i, j].is_minimizing(), True)
                self.assertEqual(model.obj3[i, j].sense, minimize)
                self.assertEqual(model.obj3[i, j].is_minimizing(), True)

    def test_dim(self):
        """Test dim method"""
        model = self.create_model()
        model.obj = Objective(model.A, model.A)

        self.assertEqual(model.obj.dim(), 2)

    def test_keys(self):
        """Test keys method"""
        model = self.create_model()

        def A_rule(model, i, j):
            return model.x

        model.x = Var()
        model.obj = Objective(model.A, model.A, rule=A_rule)

        self.assertEqual(len(list(model.obj.keys())), 4)

    def test_len(self):
        """Test len method"""
        model = self.create_model()
        model.obj = Objective(model.A, model.A)
        self.assertEqual(len(model.obj), 0)

        model = self.create_model()
        """Test rule option"""

        def f(model):
            ans = 0
            for i in model.x.keys():
                ans = ans + model.x[i]
            return ans

        model.x = Var(RangeSet(1, 4), initialize=2)
        model.obj = Objective(rule=f)

        self.assertEqual(len(model.obj), 1)


class TestObjList(unittest.TestCase):
    def create_model(self):
        model = ConcreteModel()
        model.A = Set(initialize=[1, 2, 3, 4])
        return model

    #
    # Tests that adding Objective.Skip increments
    # the internal counter but does not create an object
    #
    def test_conlist_skip(self):
        model = ConcreteModel()
        model.x = Var()
        model.c = ObjectiveList()
        self.assertTrue(1 not in model.c)
        self.assertEqual(len(model.c), 0)
        model.c.add(Objective.Skip)
        self.assertTrue(1 not in model.c)
        self.assertEqual(len(model.c), 0)
        model.c.add(model.x + 1)
        self.assertTrue(1 not in model.c)
        self.assertTrue(2 in model.c)
        self.assertEqual(len(model.c), 1)

    def test_rule_option1(self):
        """Test rule option"""
        model = self.create_model()
        model.B = RangeSet(1, 4)

        def f(model, i):
            if i > 4:
                return ObjectiveList.End
            ans = 0
            for j in model.B:
                ans = ans + model.x[j]
            ans *= i
            return ans

        model.x = Var(model.B, initialize=2)
        model.o = ObjectiveList(rule=f)

        self.assertEqual(model.o[1](), 8)
        self.assertEqual(model.o[2](), 16)
        self.assertEqual(len(model.o), 4)

    def test_rule_option2(self):
        """Test rule option"""
        model = self.create_model()
        model.B = RangeSet(1, 4)

        def f(model, i):
            if i > 2:
                return ObjectiveList.End
            i = 2 * i - 1
            ans = 0
            for j in model.B:
                ans = ans + model.x[j]
            ans *= i
            return ans

        model.x = Var(model.B, initialize=2)
        model.o = ObjectiveList(rule=f)

        self.assertEqual(model.o[1](), 8)
        self.assertEqual(len(model.o), 2)

    def test_rule_option1a(self):
        """Test rule option"""
        model = self.create_model()
        model.B = RangeSet(1, 4)

        @simple_objectivelist_rule
        def f(model, i):
            if i > 4:
                return None
            ans = 0
            for j in model.B:
                ans = ans + model.x[j]
            ans *= i
            return ans

        model.x = Var(model.B, initialize=2)
        model.o = ObjectiveList(rule=f)

        self.assertEqual(model.o[1](), 8)
        self.assertEqual(model.o[2](), 16)
        self.assertEqual(len(model.o), 4)

    def test_rule_option2a(self):
        """Test rule option"""
        model = self.create_model()
        model.B = RangeSet(1, 4)

        @simple_objectivelist_rule
        def f(model, i):
            if i > 2:
                return None
            i = 2 * i - 1
            ans = 0
            for j in model.B:
                ans = ans + model.x[j]
            ans *= i
            return ans

        model.x = Var(model.B, initialize=2)
        model.o = ObjectiveList(rule=f)

        self.assertEqual(model.o[1](), 8)
        self.assertEqual(len(model.o), 2)

    def test_rule_option3(self):
        """Test rule option"""
        model = self.create_model()
        model.y = Var(initialize=2)

        def f(model):
            yield model.y
            yield 2 * model.y
            yield 2 * model.y
            yield ObjectiveList.End

        model.c = ObjectiveList(rule=f)
        self.assertEqual(len(model.c), 3)
        self.assertEqual(model.c[1](), 2)
        model.d = ObjectiveList(rule=f(model))
        self.assertEqual(len(model.d), 3)
        self.assertEqual(model.d[1](), 2)

    def test_rule_option4(self):
        """Test rule option"""
        model = self.create_model()
        model.y = Var(initialize=2)
        model.c = ObjectiveList(rule=((i + 1) * model.y for i in range(3)))
        self.assertEqual(len(model.c), 3)
        self.assertEqual(model.c[1](), 2)

    def test_dim(self):
        """Test dim method"""
        model = self.create_model()
        model.o = ObjectiveList()

        self.assertEqual(model.o.dim(), 1)

    def test_keys(self):
        """Test keys method"""
        model = self.create_model()
        model.o = ObjectiveList()

        self.assertEqual(len(list(model.o.keys())), 0)

    def test_len(self):
        """Test len method"""
        model = self.create_model()
        model.o = ObjectiveList()

        self.assertEqual(len(model.o), 0)


class MiscObjTests(unittest.TestCase):
    def test_constructor(self):
        a = Objective(name="b")
        self.assertEqual(a.local_name, "b")
        try:
            a = Objective(foo="bar")
            self.fail("Can't specify an unexpected constructor option")
        except ValueError:
            pass

    def test_rule(self):
        def rule1(model):
            return []

        model = ConcreteModel()
        try:
            model.o = Objective(rule=rule1)
            self.fail("Error generating objective")
        except Exception:
            pass
        #
        model = ConcreteModel()

        def rule1(model):
            return 1.1

        model = ConcreteModel()
        model.o = Objective(rule=rule1)
        self.assertEqual(model.o(), 1.1)
        #
        model = ConcreteModel()

        def rule1(model, i):
            return 1.1

        model = ConcreteModel()
        model.a = Set(initialize=[1, 2, 3])
        try:
            model.o = Objective(model.a, rule=rule1)
        except Exception:
            self.fail("Error generating objective")

    def test_deprecated_rule_attribute(self):
        def rule(m):
            return m.x

        def new_rule(m):
            return -m.x

        m = ConcreteModel()
        m.x = Var()
        m.obj = Objective(rule=rule)

        self.assertIs(m.obj.rule._fcn, rule)
        with LoggingIntercept() as LOG:
            m.obj.rule = new_rule
        self.assertIn(
            "DEPRECATED: The 'Objective.rule' attribute will be made read-only",
            LOG.getvalue(),
        )
        self.assertIs(m.obj.rule, new_rule)

    def test_abstract_index(self):
        model = AbstractModel()
        model.A = Set()
        model.B = Set()
        model.C = model.A | model.B
        model.x = Objective(model.C)


if __name__ == "__main__":
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.gdp.disjunct import (
    GDP_Error,
    Disjunct,
    DisjunctData,
    Disjunction,
    DisjunctionData,
)

# Do not import these files: importing them registers the transformation
# plugins with the pyomo script so that they get automatically invoked.
# import pyomo.gdp.bigm
# import pyomo.gdp.hull



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
#
# -*- coding: utf-8 -*-
"""
Testing for the logical expression system
"""

import operator
from itertools import permutations, product

import pyomo.common.unittest as unittest

from pyomo.core.expr.cnf_walker import to_cnf
from pyomo.core.expr.sympy_tools import sympy_available
from pyomo.core.expr.visitor import identify_variables
from pyomo.environ import (
    all_different,
    count_if,
    land,
    atleast,
    atmost,
    BooleanConstant,
    BooleanVarList,
    ComponentMap,
    equivalent,
    exactly,
    implies,
    lor,
    RangeSet,
    value,
    ConcreteModel,
    BooleanVar,
    lnot,
    xor,
    Var,
    Integers,
)


def _generate_possible_truth_inputs(nargs):
    return product([True, False], repeat=nargs)


def _check_equivalent(assert_handle, expr_1, expr_2):
    expr_1_vars = list(identify_variables(expr_1, include_fixed=False))
    expr_2_vars = list(identify_variables(expr_2, include_fixed=False))
    assert_handle.assertEqual(len(expr_1_vars), len(expr_2_vars))
    for truth_combination in _generate_possible_truth_inputs(len(expr_1_vars)):
        for var, truth_value in zip(expr_1_vars, truth_combination):
            var.value = truth_value
        assert_handle.assertEqual(value(expr_1), value(expr_2))


class TestLogicalClasses(unittest.TestCase):
    def test_BooleanVar(self):
        """
        Simple construction and value setting
        """
        m = ConcreteModel()
        m.Y1 = BooleanVar()
        m.Y2 = BooleanVar()

        self.assertIsNone(m.Y1.value)
        m.Y1.set_value(False)
        self.assertFalse(m.Y1.value)
        m.Y1.set_value(True)
        self.assertTrue(m.Y1.value)

    def test_unary_not(self):
        m = ConcreteModel()
        m.Y = BooleanVar()
        op_static = lnot(m.Y)
        op_operator = ~m.Y
        for truth_combination in _generate_possible_truth_inputs(1):
            m.Y.set_value(truth_combination[0])
            correct_value = not truth_combination[0]
            self.assertEqual(value(op_static), correct_value)
            self.assertEqual(value(op_operator), correct_value)

    def test_binary_equiv(self):
        m = ConcreteModel()
        m.Y1 = BooleanVar()
        m.Y2 = BooleanVar()
        op_static = equivalent(m.Y1, m.Y2)
        op_class = m.Y1.equivalent_to(m.Y2)
        # op_operator = m.Y1 == m.Y2
        for truth_combination in _generate_possible_truth_inputs(2):
            m.Y1.value, m.Y2.value = truth_combination[0], truth_combination[1]
            correct_value = operator.eq(*truth_combination)
            self.assertEqual(value(op_static), correct_value)
            self.assertEqual(value(op_class), correct_value)
            # self.assertEqual(value(op_operator), correct_value)

    def test_binary_xor(self):
        m = ConcreteModel()
        m.Y1 = BooleanVar()
        m.Y2 = BooleanVar()
        op_static = xor(m.Y1, m.Y2)
        op_class = m.Y1.xor(m.Y2)
        op_operator = m.Y1 ^ m.Y2
        for truth_combination in _generate_possible_truth_inputs(2):
            m.Y1.value, m.Y2.value = truth_combination[0], truth_combination[1]
            correct_value = operator.xor(*truth_combination)
            self.assertEqual(value(op_static), correct_value)
            self.assertEqual(value(op_class), correct_value)
            self.assertEqual(value(op_operator), correct_value)

    def test_binary_implies(self):
        m = ConcreteModel()
        m.Y1 = BooleanVar()
        m.Y2 = BooleanVar()
        op_static = implies(m.Y1, m.Y2)
        op_class = m.Y1.implies(m.Y2)
        # op_loperator = m.Y2 << m.Y1
        # op_roperator = m.Y1 >> m.Y2
        for truth_combination in _generate_possible_truth_inputs(2):
            m.Y1.value, m.Y2.value = truth_combination[0], truth_combination[1]
            correct_value = (not truth_combination[0]) or truth_combination[1]
            self.assertEqual(value(op_static), correct_value)
            self.assertEqual(value(op_class), correct_value)
            # self.assertEqual(value(op_loperator), correct_value)
            # self.assertEqual(value(op_roperator), correct_value)
            nnf = lnot(m.Y1).lor(m.Y2)
            self.assertEqual(value(op_static), value(nnf))

    def test_binary_and(self):
        m = ConcreteModel()
        m.Y1 = BooleanVar()
        m.Y2 = BooleanVar()
        op_static = land(m.Y1, m.Y2)
        op_class = m.Y1.land(m.Y2)
        op_operator = m.Y1 & m.Y2
        for truth_combination in _generate_possible_truth_inputs(2):
            m.Y1.value, m.Y2.value = truth_combination[0], truth_combination[1]
            correct_value = all(truth_combination)
            self.assertEqual(value(op_static), correct_value)
            self.assertEqual(value(op_class), correct_value)
            self.assertEqual(value(op_operator), correct_value)

    def test_binary_or(self):
        m = ConcreteModel()
        m.Y1 = BooleanVar()
        m.Y2 = BooleanVar()
        op_static = lor(m.Y1, m.Y2)
        op_class = m.Y1.lor(m.Y2)
        op_operator = m.Y1 | m.Y2
        for truth_combination in _generate_possible_truth_inputs(2):
            m.Y1.value, m.Y2.value = truth_combination[0], truth_combination[1]
            correct_value = any(truth_combination)
            self.assertEqual(value(op_static), correct_value)
            self.assertEqual(value(op_class), correct_value)
            self.assertEqual(value(op_operator), correct_value)

    def test_nary_and(self):
        nargs = 3
        m = ConcreteModel()
        m.s = RangeSet(nargs)
        m.Y = BooleanVar(m.s)
        op_static = land(*(m.Y[i] for i in m.s))
        op_class = BooleanConstant(True)
        # op_operator = True
        for y in m.Y.values():
            op_class = op_class.land(y)
            # op_operator &= y
        for truth_combination in _generate_possible_truth_inputs(nargs):
            m.Y.set_values(dict(enumerate(truth_combination, 1)))
            correct_value = all(truth_combination)
            self.assertEqual(value(op_static), correct_value)
            self.assertEqual(value(op_class), correct_value)
            # self.assertEqual(value(op_operator), correct_value)

    def test_nary_or(self):
        nargs = 3
        m = ConcreteModel()
        m.s = RangeSet(nargs)
        m.Y = BooleanVar(m.s)
        op_static = lor(*(m.Y[i] for i in m.s))
        op_class = BooleanConstant(False)
        # op_operator = False
        for y in m.Y.values():
            op_class = op_class.lor(y)
            # op_operator |= y
        for truth_combination in _generate_possible_truth_inputs(nargs):
            m.Y.set_values(dict(enumerate(truth_combination, 1)))
            correct_value = any(truth_combination)
            self.assertEqual(value(op_static), correct_value)
            self.assertEqual(value(op_class), correct_value)
            # self.assertEqual(value(op_operator), correct_value)

    def test_nary_exactly(self):
        nargs = 5
        m = ConcreteModel()
        m.s = RangeSet(nargs)
        m.Y = BooleanVar(m.s)
        for truth_combination in _generate_possible_truth_inputs(nargs):
            for ntrue in range(nargs + 1):
                m.Y.set_values(dict(enumerate(truth_combination, 1)))
                correct_value = sum(truth_combination) == ntrue
                self.assertEqual(
                    value(exactly(ntrue, *(m.Y[i] for i in m.s))), correct_value
                )
                self.assertEqual(value(exactly(ntrue, m.Y)), correct_value)

    def test_nary_atmost(self):
        nargs = 5
        m = ConcreteModel()
        m.s = RangeSet(nargs)
        m.Y = BooleanVar(m.s)
        for truth_combination in _generate_possible_truth_inputs(nargs):
            for ntrue in range(nargs + 1):
                m.Y.set_values(dict(enumerate(truth_combination, 1)))
                correct_value = sum(truth_combination) <= ntrue
                self.assertEqual(
                    value(atmost(ntrue, *(m.Y[i] for i in m.s))), correct_value
                )
                self.assertEqual(value(atmost(ntrue, m.Y)), correct_value)

    def test_nary_atleast(self):
        nargs = 5
        m = ConcreteModel()
        m.s = RangeSet(nargs)
        m.Y = BooleanVar(m.s)
        for truth_combination in _generate_possible_truth_inputs(nargs):
            for ntrue in range(nargs + 1):
                m.Y.set_values(dict(enumerate(truth_combination, 1)))
                correct_value = sum(truth_combination) >= ntrue
                self.assertEqual(
                    value(atleast(ntrue, *(m.Y[i] for i in m.s))), correct_value
                )
                self.assertEqual(value(atleast(ntrue, m.Y)), correct_value)

    def test_nary_all_diff(self):
        m = ConcreteModel()
        m.x = Var(range(4), domain=Integers, bounds=(0, 3))
        for vals in permutations(range(4)):
            self.assertTrue(value(all_different(*vals)))
            for i, v in enumerate(vals):
                m.x[i] = v
            self.assertTrue(value(all_different(m.x)))
        self.assertFalse(value(all_different(1, 1, 2, 3)))
        m.x[0] = 1
        m.x[1] = 1
        m.x[2] = 2
        m.x[3] = 3
        self.assertFalse(value(all_different(m.x)))

    def test_count_if(self):
        nargs = 3
        m = ConcreteModel()
        m.s = RangeSet(nargs)
        m.Y = BooleanVar(m.s)
        m.x = Var(domain=Integers, bounds=(0, 3))
        for truth_combination in _generate_possible_truth_inputs(nargs):
            for ntrue in range(nargs + 1):
                m.Y.set_values(dict(enumerate(truth_combination, 1)))
                correct_value = sum(truth_combination)
                self.assertEqual(value(count_if(*(m.Y[i] for i in m.s))), correct_value)
                self.assertEqual(value(count_if(m.Y)), correct_value)
        m.x = 2
        self.assertEqual(
            value(count_if([m.Y[i] for i in m.s] + [m.x == 3])), correct_value
        )
        m.x = 3
        self.assertEqual(
            value(count_if([m.Y[i] for i in m.s] + [m.x == 3])), correct_value + 1
        )

    def test_to_string(self):
        m = ConcreteModel()
        m.Y1 = BooleanVar()
        m.Y2 = BooleanVar()
        m.Y3 = BooleanVar()
        m.Y4 = BooleanVar()
        m.int1 = Var(domain=Integers)
        m.int2 = Var(domain=Integers)

        self.assertEqual(str(land(m.Y1, m.Y2, m.Y3)), "Y1 ∧ Y2 ∧ Y3")
        self.assertEqual(str(lor(m.Y1, m.Y2, m.Y3)), "Y1 ∨ Y2 ∨ Y3")
        self.assertEqual(str(equivalent(m.Y1, m.Y2)), "Y1 iff Y2")
        self.assertEqual(str(implies(m.Y1, m.Y2)), "Y1 --> Y2")
        self.assertEqual(str(xor(m.Y1, m.Y2)), "Y1 ⊻ Y2")
        self.assertEqual(str(atleast(1, m.Y1, m.Y2)), "atleast(1: [Y1, Y2])")
        self.assertEqual(str(atmost(1, m.Y1, m.Y2)), "atmost(1: [Y1, Y2])")
        self.assertEqual(str(exactly(1, m.Y1, m.Y2)), "exactly(1: [Y1, Y2])")
        self.assertEqual(
            str(all_different(m.int1, m.int2)), "all_different(int1, int2)"
        )
        self.assertEqual(str(count_if(m.Y1, m.Y2)), "count_if(Y1, Y2)")

        # Precedence checks
        self.assertEqual(str(m.Y1.implies(m.Y2).lor(m.Y3)), "(Y1 --> Y2) ∨ Y3")
        self.assertEqual(str(m.Y1 & m.Y2 | m.Y3 ^ m.Y4), "Y1 ∧ Y2 ∨ Y3 ⊻ Y4")
        self.assertEqual(str(m.Y1 & (m.Y2 | m.Y3) ^ m.Y4), "Y1 ∧ (Y2 ∨ Y3) ⊻ Y4")
        self.assertEqual(str(m.Y1 & m.Y2 ^ m.Y3 | m.Y4), "Y1 ∧ Y2 ⊻ Y3 ∨ Y4")
        self.assertEqual(str(m.Y1 & m.Y2 ^ (m.Y3 | m.Y4)), "Y1 ∧ Y2 ⊻ (Y3 ∨ Y4)")
        self.assertEqual(str(m.Y1 & (m.Y2 ^ (m.Y3 | m.Y4))), "Y1 ∧ (Y2 ⊻ (Y3 ∨ Y4))")
        self.assertEqual(str(m.Y1 | m.Y2 ^ m.Y3 & m.Y4), "Y1 ∨ Y2 ⊻ Y3 ∧ Y4")
        self.assertEqual(str((m.Y1 | m.Y2) ^ m.Y3 & m.Y4), "(Y1 ∨ Y2) ⊻ Y3 ∧ Y4")
        self.assertEqual(str(((m.Y1 | m.Y2) ^ m.Y3) & m.Y4), "((Y1 ∨ Y2) ⊻ Y3) ∧ Y4")

    def test_node_types(self):
        m = ConcreteModel()
        m.Y1 = BooleanVar()
        m.Y2 = BooleanVar()
        m.Y3 = BooleanVar()
        m.int1 = Var(domain=Integers)
        m.int2 = Var(domain=Integers)
        m.int3 = Var(domain=Integers)

        self.assertFalse(m.Y1.is_expression_type())
        self.assertTrue(lnot(m.Y1).is_expression_type())
        self.assertTrue(equivalent(m.Y1, m.Y2).is_expression_type())
        self.assertTrue(atmost(1, [m.Y1, m.Y2, m.Y3]).is_expression_type())
        self.assertTrue(all_different(m.int1, m.int2, m.int3).is_expression_type())
        self.assertTrue(count_if(m.Y1, m.Y2, m.Y3).is_expression_type())

    def test_numeric_invalid(self):
        m = ConcreteModel()
        m.Y1 = BooleanVar()
        m.Y2 = BooleanVar()
        m.Y3 = BooleanVar()

        def iadd():
            m.Y3 += 2

        def isub():
            m.Y3 -= 2

        def imul():
            m.Y3 *= 2

        def idiv():
            m.Y3 /= 2

        def ipow():
            m.Y3 **= 2

        def iand():
            m.Y3 &= 2

        def ior():
            m.Y3 |= 2

        def ixor():
            m.Y3 ^= 2

        def invalid_expression_generator():
            yield lambda: m.Y1 + m.Y2
            yield lambda: m.Y1 - m.Y2
            yield lambda: m.Y1 * m.Y2
            yield lambda: m.Y1 / m.Y2
            yield lambda: m.Y1**m.Y2
            yield lambda: m.Y1.land(0)
            yield lambda: m.Y1.lor(0)
            yield lambda: m.Y1.xor(0)
            yield lambda: m.Y1.equivalent_to(0)
            yield lambda: m.Y1.implies(0)
            yield lambda: 0 + m.Y2
            yield lambda: 0 - m.Y2
            yield lambda: 0 * m.Y2
            yield lambda: 0 / m.Y2
            yield lambda: 0**m.Y2
            yield lambda: 0 & m.Y2
            yield lambda: 0 | m.Y2
            yield lambda: 0 ^ m.Y2
            yield lambda: m.Y3 + 2
            yield lambda: m.Y3 - 2
            yield lambda: m.Y3 * 2
            yield lambda: m.Y3 / 2
            yield lambda: m.Y3**2
            yield lambda: m.Y3 & 2
            yield lambda: m.Y3 | 2
            yield lambda: m.Y3 ^ 2
            yield iadd
            yield isub
            yield imul
            yield idiv
            yield ipow
            yield iand
            yield ior
            yield ixor

        numeric_error_msg = (
            "(?:(?:unsupported operand type)|(?:operands do not support))"
        )
        for i, invalid_expr_fcn in enumerate(invalid_expression_generator()):
            with self.assertRaisesRegex(TypeError, numeric_error_msg):
                _ = invalid_expr_fcn()

        def invalid_unary_expression_generator():
            yield lambda: -m.Y1
            yield lambda: +m.Y1

        for invalid_expr_fcn in invalid_unary_expression_generator():
            with self.assertRaisesRegex(
                TypeError,
                "(?:(?:bad operand type for unary)"
                "|(?:unsupported operand type for unary))",
            ):
                _ = invalid_expr_fcn()

        def invalid_comparison_generator():
            yield lambda: m.Y1 >= 0
            yield lambda: m.Y1 <= 0
            yield lambda: m.Y1 > 0
            yield lambda: m.Y1 < 0

        # These errors differ between python versions, regrettably
        comparison_error_msg = (
            "(?:(?:unorderable types)|(?:not supported between instances of))"
        )
        for invalid_expr_fcn in invalid_comparison_generator():
            with self.assertRaisesRegex(TypeError, comparison_error_msg):
                _ = invalid_expr_fcn()

    def test_invalid_conversion(self):
        m = ConcreteModel()
        m.Y1 = BooleanVar()

        with self.assertRaisesRegex(
            TypeError, r"argument must be a string or a(.*) number"
        ):
            float(m.Y1)

        with self.assertRaisesRegex(
            TypeError,
            r"argument must be a string" r"(?:, a bytes-like object)? or a(.*) number",
        ):
            int(m.Y1)


@unittest.skipUnless(sympy_available, "sympy not available")
class TestCNF(unittest.TestCase):
    def test_cnf(self):
        m = ConcreteModel()
        m.Y1 = BooleanVar()
        m.Y2 = BooleanVar()

        implication = implies(m.Y1, m.Y2)
        x = to_cnf(implication)[0]
        _check_equivalent(self, implication, x)

        atleast_expr = atleast(1, m.Y1, m.Y2)
        x = to_cnf(atleast_expr)[0]
        self.assertIs(atleast_expr, x)  # should be no change

        nestedatleast = implies(m.Y1, atleast_expr)
        m.extraY = BooleanVarList()
        indicator_map = ComponentMap()
        x = to_cnf(nestedatleast, m.extraY, indicator_map)
        self.assertEqual(str(x[0]), "extraY[1] ∨ ~Y1")
        self.assertIs(indicator_map[m.extraY[1]], atleast_expr)

    # TODO need to test other combinations as well


if __name__ == "__main__":
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

#
# Test different transformations for complementarity conditions.
#
# These perform difference tests with files that capture the structure of
# the transformed model.
#

import os

from filecmp import cmp
import pyomo.common.unittest as unittest

from pyomo.common.fileutils import this_file_dir
from pyomo.common.tee import capture_output
from pyomo.common.tempfiles import TempfileManager
from pyomo.core import (
    ConcreteModel,
    Var,
    Constraint,
    TransformationFactory,
    Objective,
    Block,
    inequality,
)
from pyomo.gdp import Disjunct, Disjunction
from pyomo.mpec import Complementarity, complements, ComplementarityList
from pyomo.opt import ProblemFormat
from pyomo.repn.plugins.nl_writer import FileDeterminism
from pyomo.repn.tests.nl_diff import load_and_compare_nl_baseline

currdir = this_file_dir()


class CCTests:
    @classmethod
    def setUpClass(cls):
        import pyomo.environ

    def _setup(self):
        M = ConcreteModel()
        M.y = Var()
        M.x1 = Var()
        M.x2 = Var()
        M.x3 = Var()
        return M

    def _print(self, model):
        model.cc.pprint()

    def _test(self, tname, M):
        bfile = os.path.join(currdir, tname + f'_{self.xfrm}.txt')
        if self.xfrm is not None:
            xfrm = TransformationFactory(self.xfrm)
            xfrm.apply_to(M)
        with TempfileManager:
            ofile = TempfileManager.create_tempfile(suffix=f'_{self.xfrm}.out')
            with capture_output(ofile):
                self._print(M)
            try:
                self.assertTrue(
                    cmp(ofile, bfile), msg="Files %s and %s differ" % (ofile, bfile)
                )
            except:
                with open(ofile, 'r') as f1, open(bfile, 'r') as f2:
                    f1_contents = list(filter(None, f1.read().split()))
                    f2_contents = list(filter(None, f2.read().split()))
                    self.assertEqual(f1_contents, f2_contents)

    def test_t1a(self):
        # y + x1 >= 0  _|_  x1 + 2*x2 + 3*x3 >= 1
        M = self._setup()
        M.c = Constraint(expr=M.y + M.x3 >= M.x2)
        M.cc = Complementarity(
            expr=complements(M.y + M.x1 >= 0, M.x1 + 2 * M.x2 + 3 * M.x3 >= 1)
        )
        self._test("t1a", M)

    def test_t1b(self):
        # Reversing the expressions in test t1a:
        #    x1 + 2*x2 + 3*x3 >= 1  _|_  y + x1 >= 0
        M = self._setup()
        M.cc = Complementarity(
            expr=complements(M.x1 + 2 * M.x2 + 3 * M.x3 >= 1, M.y + M.x1 >= 0)
        )
        self._test("t1b", M)

    def test_t1c(self):
        # y >= - x1  _|_  x1 + 2*x2 >= 1 - 3*x3
        M = self._setup()
        M.cc = Complementarity(
            expr=complements(M.y >= -M.x1, M.x1 + 2 * M.x2 >= 1 - 3 * M.x3)
        )
        self._test("t1c", M)

    def test_t2a(self):
        # y + x2 >= 0  _|_  x2 - x3 <= -1
        M = self._setup()
        M.cc = Complementarity(expr=complements(M.y + M.x2 >= 0, M.x2 - M.x3 <= -1))
        self._test("t2a", M)

    def test_t2b(self):
        # Reversing the expressions in test t2a:
        #    x2 - x3 <= -1  _|_  y + x2 >= 0
        M = self._setup()
        M.cc = Complementarity(expr=complements(M.x2 - M.x3 <= -1, M.y + M.x2 >= 0))
        self._test("t2b", M)

    def test_t3a(self):
        # y + x3 >= 0  _|_  x1 + x2 >= -1
        M = self._setup()
        M.cc = Complementarity(expr=complements(M.y + M.x3 >= 0, M.x1 + M.x2 >= -1))
        self._test("t3a", M)

    def test_t3b(self):
        # Reversing the expressions in test t3a:
        #    x1 + x2 >= -1  _|_  y + x3 >= 0
        M = self._setup()
        M.cc = Complementarity(expr=complements(M.x1 + M.x2 >= -1, M.y + M.x3 >= 0))
        self._test("t3b", M)

    def test_t4a(self):
        # x1 + 2*x2 + 3*x3 = 1  _|_  y + x3
        M = self._setup()
        M.cc = Complementarity(
            expr=complements(M.x1 + 2 * M.x2 + 3 * M.x3 == 1, M.y + M.x3)
        )
        self._test("t4a", M)

    def test_t4b(self):
        # Reversing the expressions in test t7b:
        #    y + x3  _|_  x1 + 2*x2 + 3*x3 = 1
        M = self._setup()
        M.cc = Complementarity(
            expr=complements(M.y + M.x3, M.x1 + 2 * M.x2 + 3 * M.x3 == 1)
        )
        self._test("t4b", M)

    def test_t4c(self):
        # 1 = x1 + 2*x2 + 3*x3  _|_  y + x3
        M = self._setup()
        M.cc = Complementarity(
            expr=complements(1 == M.x1 + 2 * M.x2 + 3 * M.x3, M.y + M.x3)
        )
        self._test("t4c", M)

    def test_t4d(self):
        # x1 + 2*x2 == 1 - 3*x3  _|_  y + x3
        M = self._setup()
        M.cc = Complementarity(
            expr=complements(M.x1 + 2 * M.x2 == 1 - 3 * M.x3, M.y + M.x3)
        )
        self._test("t4d", M)

    def test_t9(self):
        # Testing that we can skip deactivated complementarity conditions
        M = self._setup()
        M.cc = Complementarity(expr=complements(M.y + M.x3, M.x1 + 2 * M.x2 == 1))
        M.cc.deactivate()
        # AMPL needs at least one variable in the problem therefore
        # we need to have a constraint that keeps them around
        M.keep_var_con = Constraint(expr=M.x1 == 0.5)
        self._test("t9", M)

    def test_t10(self):
        # Testing that we can skip an array of deactivated complementarity conditions
        M = self._setup()

        def f(model, i):
            return complements(M.y + M.x3, M.x1 + 2 * M.x2 == i)

        M.cc = Complementarity([0, 1, 2], rule=f)
        M.cc[1].deactivate()
        self._test("t10", M)

    def test_t11(self):
        # 2 <= y + x1 <= 3  _|_  x1
        M = self._setup()
        M.cc = Complementarity(expr=complements(inequality(2, M.y + M.x1, 3), M.x1))
        self._test("t11", M)

    def test_t12(self):
        # x1  _|_  2 <= y + x1 <= 3"""
        M = self._setup()
        M.cc = Complementarity(expr=complements(M.x1, inequality(2, M.y + M.x1, 3)))
        self._test("t12", M)

    def test_t13(self):
        # Testing that we can skip an array of deactivated complementarity conditions
        M = self._setup()

        def f(model, i):
            if i == 0:
                return complements(M.y + M.x3, M.x1 + 2 * M.x2 == 0)
            if i == 1:
                return Complementarity.Skip
            if i == 2:
                return complements(M.y + M.x3, M.x1 + 2 * M.x2 == 2)

        M.cc = Complementarity([0, 1, 2], rule=f)
        self._test("t13", M)

    def test_cov2(self):
        # Testing warning for no rule"""
        M = self._setup()
        M.cc = Complementarity([0, 1, 2])
        # AMPL needs at least one variable in the problem therefore
        # we need to have a constraint that keeps them around
        M.keep_var_con = Constraint(expr=M.x1 == 0.5)
        self._test("cov2", M)

    def test_cov4(self):
        # Testing construction with no indexing and a rule
        M = self._setup()

        def f(model):
            return complements(M.y + M.x3, M.x1 + 2 * M.x2 == 1)

        M.cc = Complementarity(rule=f)
        self._test("cov4", M)

    def test_cov5(self):
        # Testing construction with rules that generate an exception
        M = self._setup()

        def f(model):
            raise IOError("cov5 error")

        try:
            M.cc1 = Complementarity(rule=f)
            self.fail("Expected an IOError")
        except IOError:
            pass

        def f(model, i):
            raise IOError("cov5 error")

        try:
            M.cc2 = Complementarity([0, 1], rule=f)
            self.fail("Expected an IOError")
        except IOError:
            pass

    def test_cov6(self):
        # Testing construction with indexing and an expression
        M = self._setup()
        with self.assertRaisesRegex(ValueError, "Invalid tuple for Complementarity"):
            M.cc = Complementarity([0, 1], expr=())

    def test_cov7(self):
        # Testing error checking with return value
        M = self._setup()

        def f(model):
            return ()

        try:
            M.cc = Complementarity(rule=f)
            self.fail("Expected ValueError")
        except ValueError:
            pass

        def f(model):
            return

        try:
            M.cc = Complementarity(rule=f)
            self.fail("Expected ValueError")
        except ValueError:
            pass

        def f(model):
            return {}

        try:
            M.cc = Complementarity(rule=f)
            self.fail("Expected ValueError")
        except ValueError:
            pass

    def test_cov8(self):
        # Testing construction with a list
        M = self._setup()

        def f(model):
            return [M.y + M.x3, M.x1 + 2 * M.x2 == 1]

        M.cc = Complementarity(rule=f)
        self._test("cov8", M)

    def test_cov9(self):
        # Testing construction with a tuple
        M = self._setup()

        def f(model):
            return (M.y + M.x3, M.x1 + 2 * M.x2 == 1)

        M.cc = Complementarity(rule=f)
        self._test("cov8", M)

    def test_cov10(self):
        # Testing construction with a badly formed expression
        M = self._setup()
        M.cc = Complementarity(expr=complements(inequality(M.y, M.x1, 1), M.x2))
        try:
            M.cc.to_standard_form()
            self.fail("Expected a RuntimeError")
        except RuntimeError:
            pass

    def test_cov11(self):
        # Testing construction with a badly formed expression
        M = self._setup()
        M.cc = Complementarity(expr=complements(inequality(1, M.x1, M.y), M.x2))
        try:
            M.cc.to_standard_form()
            self.fail("Expected a RuntimeError")
        except RuntimeError:
            pass

    def test_list1(self):
        M = self._setup()
        M.cc = ComplementarityList()
        M.cc.add(complements(M.y + M.x3, M.x1 + 2 * M.x2 == 0))
        M.cc.add(complements(M.y + M.x3, M.x1 + 2 * M.x2 == 2))
        self._test("list1", M)

    def test_list2(self):
        M = self._setup()
        M.cc = ComplementarityList()
        M.cc.add(complements(M.y + M.x3, M.x1 + 2 * M.x2 == 0))
        M.cc.add(complements(M.y + M.x3, M.x1 + 2 * M.x2 == 1))
        M.cc.add(complements(M.y + M.x3, M.x1 + 2 * M.x2 == 2))
        M.cc[2].deactivate()
        self._test("list2", M)

    def test_list3(self):
        M = self._setup()

        def f(M, i):
            if i == 1:
                return complements(M.y + M.x3, M.x1 + 2 * M.x2 == 0)
            elif i == 2:
                return complements(M.y + M.x3, M.x1 + 2 * M.x2 == 2)
            return ComplementarityList.End

        M.cc = ComplementarityList(rule=f)
        self._test("list1", M)

    def test_list4(self):
        M = self._setup()

        def f(M):
            yield complements(M.y + M.x3, M.x1 + 2 * M.x2 == 0)
            yield complements(M.y + M.x3, M.x1 + 2 * M.x2 == 2)
            yield ComplementarityList.End

        M.cc = ComplementarityList(rule=f)
        self._test("list1", M)

    def test_list5(self):
        M = self._setup()
        M.cc = ComplementarityList(
            rule=(complements(M.y + M.x3, M.x1 + 2 * M.x2 == i) for i in range(3))
        )
        self._test("list5", M)

    def test_list6(self):
        M = self._setup()
        try:
            M.cc = ComplementarityList()
            self.fail("Expected a RuntimeError")
        except:
            pass

    def test_list7(self):
        M = self._setup()

        def f(M):
            return None

        try:
            M.cc = ComplementarityList(rule=f)
            self.fail("Expected a ValueError")
        except:
            pass
        M = self._setup()

        def f(M):
            yield None

        try:
            M.cc = ComplementarityList(rule=f)
            self.fail("Expected a ValueError")
        except:
            pass


class CCTests_none(CCTests, unittest.TestCase):
    xfrm = None


class CCTests_nl(CCTests, unittest.TestCase):
    xfrm = 'mpec.nl'

    def _print(self, model):
        model.pprint()


class CCTests_standard_form(CCTests, unittest.TestCase):
    xfrm = 'mpec.standard_form'


class CCTests_simple_nonlinear(CCTests, unittest.TestCase):
    xfrm = 'mpec.simple_nonlinear'


class CCTests_simple_disjunction(CCTests, unittest.TestCase):
    xfrm = 'mpec.simple_disjunction'


class CCTests_nl_nlxfrm(CCTests):
    def _test(self, tname, M):
        bfile = os.path.join(currdir, tname + '_nlxfrm.nl')
        xfrm = TransformationFactory('mpec.nl')
        xfrm.apply_to(M)
        fd = FileDeterminism.SORT_INDICES if self._nl_version == 'nl_v2' else 1
        with TempfileManager:
            ofile = TempfileManager.create_tempfile(suffix='_nlxfrm.out')
            M.write(
                ofile,
                format=self._nl_version,
                io_options={'symbolic_solver_labels': False, 'file_determinism': fd},
            )
            self.assertEqual(
                *load_and_compare_nl_baseline(bfile, ofile, self._nl_version)
            )


class CCTests_nl_nlxfrm_nlv1(CCTests_nl_nlxfrm, unittest.TestCase):
    _nl_version = 'nl_v1'


class CCTests_nl_nlxfrm_nlv2(CCTests_nl_nlxfrm, unittest.TestCase):
    _nl_version = 'nl_v2'


class DescendIntoDisjunct(unittest.TestCase):
    def get_model(self):
        m = ConcreteModel()
        m.x = Var(bounds=(-100, 100))

        m.obj = Objective(expr=m.x)

        m.disjunct1 = Disjunct()
        m.disjunct1.comp = Complementarity(expr=complements(m.x >= 0, 4 * m.x - 3 >= 0))
        m.disjunct2 = Disjunct()
        m.disjunct2.cons = Constraint(expr=m.x >= 2)

        m.disjunction = Disjunction(expr=[m.disjunct1, m.disjunct2])

        return m

    def check_simple_disjunction(self, m):
        # check that we have what we expect on disjunct1
        compBlock = m.disjunct1.component('comp')
        self.assertIsInstance(compBlock, Block)
        self.assertIsInstance(compBlock.component('expr1'), Disjunct)
        self.assertIsInstance(compBlock.component('expr2'), Disjunct)
        self.assertIsInstance(compBlock.component('complements'), Disjunction)

    def test_simple_disjunction_descend_into_disjunct(self):
        m = self.get_model()
        TransformationFactory('mpec.simple_disjunction').apply_to(m)
        self.check_simple_disjunction(m)

    def test_simple_disjunction_on_disjunct(self):
        m = self.get_model()
        TransformationFactory('mpec.simple_disjunction').apply_to(m.disjunct1)
        self.check_simple_disjunction(m)

    def check_simple_nonlinear(self, m):
        # check that we have what we expect on disjunct1
        compBlock = m.disjunct1.component('comp')
        self.assertIsInstance(compBlock, Block)
        self.assertIsInstance(compBlock.component('v'), Var)
        self.assertIsInstance(compBlock.component('c'), Constraint)
        self.assertIsInstance(compBlock.component('ccon'), Constraint)
        self.assertIsInstance(compBlock.component('ve'), Constraint)

    def test_simple_nonlinear_descend_into_disjunct(self):
        m = self.get_model()
        TransformationFactory('mpec.simple_nonlinear').apply_to(m)
        self.check_simple_nonlinear(m)

    def test_simple_nonlinear_on_disjunct(self):
        m = self.get_model()

    def check_standard_form(self, m):
        # check that we have what we expect on disjunct1
        compBlock = m.disjunct1.component('comp')
        self.assertIsInstance(compBlock, Block)
        self.assertIsInstance(compBlock.component('v'), Var)
        self.assertIsInstance(compBlock.component('c'), Constraint)
        self.assertIsInstance(compBlock.component('ve'), Constraint)

    def test_standard_form_descend_into_disjunct(self):
        m = self.get_model()
        TransformationFactory('mpec.standard_form').apply_to(m)
        self.check_standard_form(m)

    def test_standard_form_on_disjunct(self):
        m = self.get_model()
        TransformationFactory('mpec.standard_form').apply_to(m.disjunct1)
        self.check_standard_form(m)

    def check_nl(self, m):
        compBlock = m.disjunct1.component('comp')
        self.assertIsInstance(compBlock, Block)
        self.assertIsInstance(compBlock.component('bv'), Var)
        self.assertIsInstance(compBlock.component('c'), Constraint)
        self.assertIsInstance(compBlock.component('bc'), Constraint)

    def test_nl_descend_into_disjunct(self):
        m = self.get_model()
        TransformationFactory('mpec.nl').apply_to(m)
        self.check_nl(m)

    def test_nl_on_disjunct(self):
        m = self.get_model()
        TransformationFactory('mpec.nl').apply_to(m.disjunct1)
        self.check_nl(m)


if __name__ == "__main__":
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import pyomo.kernel as pmo
from pyomo.core import ConcreteModel, Var, Objective, Constraint, Binary, maximize
from pyomo.solvers.tests.models.base import _BaseTestModel, register_model


@register_model
class MIQCP_simple(_BaseTestModel):
    """
    A mixed-integer model with a quadratic objective and quadratic constraints
    """

    description = "MIQCP_simple"
    level = ('nightly', 'expensive')
    capabilities = set(['linear', 'integer', 'quadratic_constraint'])

    def __init__(self):
        _BaseTestModel.__init__(self)
        self.add_results(self.description + ".json")

    def _generate_model(self):
        self.model = ConcreteModel()
        model = self.model
        model._name = self.description

        model.x = Var(within=Binary)
        model.y = Var(within=Binary)
        model.z = Var(within=Binary)

        model.obj = Objective(expr=model.x, sense=maximize)
        model.c0 = Constraint(expr=model.x + model.y + model.z == 1)
        model.qc0 = Constraint(expr=model.x**2 + model.y**2 <= model.z**2)
        model.qc1 = Constraint(expr=model.x**2 <= model.y * model.z)

    def warmstart_model(self):
        assert self.model is not None
        model = self.model
        model.x.value = None
        model.y.value = None
        model.z.value = None


@register_model
class MIQCP_simple_kernel(MIQCP_simple):
    def _generate_model(self):
        self.model = pmo.block()
        model = self.model
        model._name = self.description

        model.x = pmo.variable(domain=Binary)
        model.y = pmo.variable(domain=Binary)
        model.z = pmo.variable(domain=Binary)

        model.obj = pmo.objective(model.x, sense=maximize)
        model.c0 = pmo.constraint(model.x + model.y + model.z == 1)
        model.qc0 = pmo.constraint(model.x**2 + model.y**2 <= model.z**2)
        model.qc1 = pmo.constraint(model.x**2 <= model.y * model.z)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.common.dependencies import pandas as pd
import pyomo.contrib.parmest.parmest as parmest
from pyomo.contrib.parmest.examples.rooney_biegler.rooney_biegler import (
    RooneyBieglerExperiment,
)


def main():

    # Data
    data = pd.DataFrame(
        data=[[1, 8.3], [2, 10.3], [3, 19.0], [4, 16.0], [5, 15.6], [7, 19.8]],
        columns=['hour', 'y'],
    )

    # Create an experiment list
    exp_list = []
    for i in range(data.shape[0]):
        exp_list.append(RooneyBieglerExperiment(data.loc[i, :]))

    # Create an instance of the parmest estimator
    pest = parmest.Estimator(exp_list, obj_function="SSE")

    # Parameter estimation and covariance
    obj, theta = pest.theta_est()
    cov = pest.cov_est()

    if parmest.graphics.seaborn_available:
        parmest.graphics.pairwise_plot(
            (theta, cov, 100),
            theta_star=theta,
            alpha=0.8,
            distributions=['MVN'],
            title='Theta estimates within 80% confidence region',
        )

    # Assert statements compare parameter estimation (theta) to an expected value
    relative_error = abs(theta['asymptote'] - 19.1426) / 19.1426
    assert relative_error < 0.01
    relative_error = abs(theta['rate_constant'] - 0.5311) / 0.5311
    assert relative_error < 0.01

    return obj, theta, cov


if __name__ == "__main__":
    obj, theta, cov = main()
    print("Estimated parameters (theta):", theta)
    print("Objective function value at theta:", obj)
    print("Covariance of parameter estimates:", cov)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import pickle

import pyomo.common.unittest as unittest
from pyomo.core.tests.unit.kernel.test_dict_container import (
    _TestActiveDictContainerBase,
)
from pyomo.core.tests.unit.kernel.test_tuple_container import (
    _TestActiveTupleContainerBase,
)
from pyomo.core.tests.unit.kernel.test_list_container import (
    _TestActiveListContainerBase,
)
from pyomo.core.kernel.base import ICategorizedObject
from pyomo.core.kernel.sos import ISOS, sos, sos1, sos2, sos_dict, sos_tuple, sos_list
from pyomo.core.kernel.block import block
from pyomo.core.kernel.variable import variable, variable_list
from pyomo.core.kernel.parameter import parameter
from pyomo.core.kernel.expression import expression, data_expression


class Test_sos(unittest.TestCase):
    def test_pprint(self):
        import pyomo.kernel

        # Not really testing what the output is, just that
        # an error does not occur. The pprint functionality
        # is still in the early stages.
        vlist = variable_list([variable(), variable()])
        s = sos(vlist)
        pyomo.kernel.pprint(s)
        b = block()
        b.s = s
        pyomo.kernel.pprint(s)
        pyomo.kernel.pprint(b)
        m = block()
        m.b = b
        pyomo.kernel.pprint(s)
        pyomo.kernel.pprint(b)
        pyomo.kernel.pprint(m)

    def test_ctype(self):
        s = sos([])
        self.assertIs(s.ctype, ISOS)
        self.assertIs(type(s), sos)
        self.assertIs(type(s)._ctype, ISOS)

    def test_pickle(self):
        v = variable()
        s = sos([v], weights=[1])
        self.assertEqual(len(s), 1)
        self.assertIs(s.variables[0], v)
        self.assertTrue(v in s)
        self.assertEqual(s.weights[0], 1)
        self.assertEqual(s.level, 1)
        self.assertEqual(s.parent, None)
        sup = pickle.loads(pickle.dumps(s))
        self.assertEqual(len(sup), 1)
        self.assertIsNot(sup.variables[0], v)
        self.assertFalse(v in sup)
        self.assertEqual(sup.weights[0], 1)
        self.assertEqual(sup.level, 1)
        self.assertEqual(sup.parent, None)

        b = block()
        b.v = v
        self.assertIs(v.parent, b)
        b.s = s
        self.assertIs(s.parent, b)
        bup = pickle.loads(pickle.dumps(b))
        sup = bup.s
        self.assertEqual(len(sup), 1)
        self.assertIs(sup.variables[0], bup.v)
        self.assertTrue(bup.v in sup)
        self.assertEqual(sup.weights[0], 1)
        self.assertEqual(sup.level, 1)
        self.assertIs(sup.parent, bup)

    def test_init(self):
        s = sos([])
        self.assertTrue(s.parent is None)
        self.assertEqual(s.ctype, ISOS)
        self.assertEqual(len(s), 0)
        self.assertEqual(s.variables, ())
        self.assertEqual(s.weights, ())
        self.assertEqual(s.level, 1)

        vlist = tuple([variable(), variable()])
        s = sos(vlist)
        self.assertTrue(s.parent is None)
        self.assertEqual(s.ctype, ISOS)
        self.assertEqual(len(s), 2)
        self.assertEqual(len(s.variables), 2)
        for v in vlist:
            self.assertTrue(v in s)
        self.assertEqual(s.weights, tuple([1, 2]))
        self.assertEqual(s.level, 1)

        vlist = tuple([variable(), variable()])
        s = sos(vlist, weights=[3.5, 4.5], level=2)
        self.assertTrue(s.parent is None)
        self.assertEqual(s.ctype, ISOS)
        self.assertEqual(len(s), 2)
        self.assertEqual(len(s.variables), 2)
        for v in vlist:
            self.assertTrue(v in s)
        self.assertEqual(s.weights, tuple([3.5, 4.5]))
        self.assertEqual(s.level, 2)
        for i, (v, w) in enumerate(s.items()):
            self.assertIs(v, vlist[i])
            self.assertEqual(w, s.weights[i])

    def test_type(self):
        s = sos([])
        self.assertTrue(isinstance(s, ICategorizedObject))
        self.assertTrue(isinstance(s, ISOS))

        s = sos1([])
        self.assertTrue(isinstance(s, ICategorizedObject))
        self.assertTrue(isinstance(s, ISOS))

        s = sos2([])
        self.assertTrue(isinstance(s, ICategorizedObject))
        self.assertTrue(isinstance(s, ISOS))

    def test_bad_weights(self):
        v = variable()
        with self.assertRaises(ValueError):
            s = sos([v], weights=[v])

        v.fix(1.0)
        with self.assertRaises(ValueError):
            s = sos([v], weights=[v])

        e = expression()
        with self.assertRaises(ValueError):
            s = sos([v], weights=[e])

        de = data_expression()
        s = sos([v], weights=[de])

        p = parameter()
        s = sos([v], weights=[p])

    def test_active(self):
        s = sos([])
        self.assertEqual(s.active, True)
        s.deactivate()
        self.assertEqual(s.active, False)
        s.activate()
        self.assertEqual(s.active, True)

        b = block()
        self.assertEqual(b.active, True)
        b.deactivate()
        self.assertEqual(b.active, False)
        b.s = s
        self.assertEqual(s.active, True)
        self.assertEqual(b.active, False)
        s.deactivate()
        self.assertEqual(s.active, False)
        self.assertEqual(b.active, False)
        b.activate()
        self.assertEqual(s.active, False)
        self.assertEqual(b.active, True)
        b.activate(shallow=False)
        self.assertEqual(s.active, True)
        self.assertEqual(b.active, True)
        b.deactivate(shallow=False)
        self.assertEqual(s.active, False)
        self.assertEqual(b.active, False)


class Test_sos_dict(_TestActiveDictContainerBase, unittest.TestCase):
    _container_type = sos_dict
    _ctype_factory = lambda self: sos([variable()])


class Test_sos_tuple(_TestActiveTupleContainerBase, unittest.TestCase):
    _container_type = sos_tuple
    _ctype_factory = lambda self: sos([variable()])


class Test_sos_list(_TestActiveListContainerBase, unittest.TestCase):
    _container_type = sos_list
    _ctype_factory = lambda self: sos([variable()])


if __name__ == "__main__":
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

# -*- coding: utf-8 -*-
"""Example 1 in Paper 'Using regularization and second order information in outer approximation for convex MINLP'

The expected optimal solution value is -56.981.

Ref:
    Kronqvist J, Bernal D E, Grossmann I E. Using regularization and second order information in outer approximation for convex MINLP[J]. Mathematical Programming, 2020, 180(1): 285-310.

    Problem type:    convex MINLP
            size:    1  binary variable
                     1  continuous variables
                     3  constraints


"""

from pyomo.environ import (
    ConcreteModel,
    Constraint,
    Reals,
    Integers,
    Objective,
    Var,
    sqrt,
    minimize,
    exp,
)
from pyomo.common.collections import ComponentMap


class SimpleMINLP4(ConcreteModel):
    def __init__(self, *args, **kwargs):
        """Create the problem."""
        kwargs.setdefault('name', 'SimpleMINLP4')
        super(SimpleMINLP4, self).__init__(*args, **kwargs)
        m = self

        m.x = Var(domain=Reals, bounds=(1, 20), initialize=5.29)
        m.y = Var(domain=Integers, bounds=(1, 20), initialize=3)

        m.c1 = Constraint(
            expr=0.3 * (m.x - 8) ** 2
            + 0.04 * (m.y - 6) ** 4
            + 0.1 * exp(2 * m.x) * ((m.y) ** (-4))
            <= 56
        )
        m.c2 = Constraint(expr=1 / m.x + 1 / m.y - sqrt(m.x) * sqrt(m.y) <= -1)
        m.c3 = Constraint(expr=2 * m.x - 5 * m.y <= -1)

        m.objective = Objective(expr=-6 * m.x - m.y, sense=minimize)
        m.optimal_value = -56.981
        m.optimal_solution = ComponentMap()
        m.optimal_solution[m.x] = 7.663528589138092
        m.optimal_solution[m.y] = 11.0



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from . import common
from .version import __version__

#
# declare deprecation paths for removed modules
#
from pyomo.common.deprecation import moved_module

moved_module(
    'pyomo.pysp',
    'pysp',
    version='6.0',
    msg="PySP has been removed from the pyomo.pysp namespace.  "
    "Beginning in Pyomo 6.0, PySP is distributed as a separate "
    "package.  Please see https://github.com/Pyomo/pysp for "
    "information on downloading and installing PySP",
)
# (*silently*) redirect 'pyomo.__future__' to 'pyomo.future'
moved_module('pyomo.__future__', 'pyomo.future', msg=None)
del moved_module



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import logging

from pyomo.gdp import GDP_Error
from pyomo.common.collections import ComponentSet
from pyomo.contrib.fbbt.expression_bounds_walker import ExpressionBoundsVisitor
import pyomo.contrib.fbbt.interval as interval
from pyomo.core import Suffix

logger = logging.getLogger(__name__)


def _convert_M_to_tuple(M, constraint, disjunct=None):
    if not isinstance(M, (tuple, list)):
        if M is None:
            M = (None, None)
        else:
            try:
                M = (-M, M)
            except Exception:
                logger.error(
                    "Error converting scalar M-value %s "
                    "to (-M,M).  Is %s not a numeric type?" % (M, type(M))
                )
                raise
    if len(M) != 2:
        constraint_name = constraint.name
        if disjunct is not None:
            constraint_name += " relative to Disjunct %s" % disjunct.name
        raise GDP_Error(
            "Big-M %s for constraint %s is not of "
            "length two. "
            "Expected either a single value or "
            "tuple or list of length two specifying M values for "
            "the lower and upper sides of the constraint "
            "respectively." % (str(M), constraint.name)
        )

    return M


def _get_bigM_suffix_list(block, stopping_block=None):
    # Note that you can only specify suffixes on BlockData objects or
    # ScalarBlocks. Though it is possible at this point to stick them
    # on whatever components you want, we won't pick them up.
    suffix_list = []

    # go searching above block in the tree, stop when we hit stopping_block
    # (This is so that we can search on each Disjunct once, but get any
    # information between a constraint and its Disjunct while transforming
    # the constraint).
    while block is not None:
        bigm = block.component('BigM')
        if type(bigm) is Suffix:
            suffix_list.append(bigm)
        if block is stopping_block:
            break
        block = block.parent_block()

    return suffix_list


def _warn_for_unused_bigM_args(bigM, used_args, logger):
    # issue warnings about anything that was in the bigM args dict that we
    # didn't use
    if bigM is not None:
        unused_args = ComponentSet(bigM.keys()) - ComponentSet(used_args.keys())
        if len(unused_args) > 0:
            warning_msg = (
                "Unused arguments in the bigM map! "
                "These arguments were not used by the "
                "transformation:\n"
            )
            for component in unused_args:
                if isinstance(component, (tuple, list)) and len(component) == 2:
                    warning_msg += "\t(%s, %s)\n" % (
                        component[0].name,
                        component[1].name,
                    )
                elif hasattr(component, 'name'):
                    warning_msg += "\t%s\n" % component.name
                else:
                    warning_msg += "\t%s\n" % component
            logger.warning(warning_msg)


class _BigM_MixIn:
    def _get_bigM_arg_list(self, bigm_args, block):
        # Gather what we know about blocks from args exactly once. We'll still
        # check for constraints in the moment, but if that fails, we've
        # preprocessed the time-consuming part of traversing up the tree.
        arg_list = []
        if bigm_args is None:
            return arg_list
        while block is not None:
            if block in bigm_args:
                arg_list.append({block: bigm_args[block]})
            block = block.parent_block()
        return arg_list

    def _set_up_expr_bound_visitor(self):
        # we assume the default config arg for 'assume_fixed_vars_permanent,`
        # and we will change it during apply_to if we need to
        self._expr_bound_visitor = ExpressionBoundsVisitor(
            use_fixed_var_values_as_bounds=False
        )

    def _process_M_value(
        self,
        m,
        lower,
        upper,
        need_lower,
        need_upper,
        src,
        key,
        constraint,
        from_args=False,
    ):
        m = _convert_M_to_tuple(m, constraint)
        if need_lower and m[0] is not None:
            if from_args:
                self.used_args[key] = m
            lower = (m[0], src, key)
            need_lower = False
        if need_upper and m[1] is not None:
            if from_args:
                self.used_args[key] = m
            upper = (m[1], src, key)
            need_upper = False
        return lower, upper, need_lower, need_upper

    def _get_M_from_args(self, constraint, bigMargs, arg_list, lower, upper):
        # check args: we first look in the keys for constraint and
        # constraintdata. In the absence of those, we traverse up the blocks,
        # and as a last resort check for a value for None
        if bigMargs is None:
            return (lower, upper)

        # since we check for args first, we know lower[0] and upper[0] are both
        # None
        need_lower = constraint.lower is not None
        need_upper = constraint.upper is not None

        # check for the constraint itself and its container
        parent = constraint.parent_component()
        if constraint in bigMargs:
            m = bigMargs[constraint]
            lower, upper, need_lower, need_upper = self._process_M_value(
                m,
                lower,
                upper,
                need_lower,
                need_upper,
                bigMargs,
                constraint,
                constraint,
                from_args=True,
            )
            if not need_lower and not need_upper:
                return lower, upper
        elif parent in bigMargs:
            m = bigMargs[parent]
            lower, upper, need_lower, need_upper = self._process_M_value(
                m,
                lower,
                upper,
                need_lower,
                need_upper,
                bigMargs,
                parent,
                constraint,
                from_args=True,
            )
            if not need_lower and not need_upper:
                return lower, upper

        # use the precomputed traversal up the blocks
        for arg in arg_list:
            for block, val in arg.items():
                lower, upper, need_lower, need_upper = self._process_M_value(
                    val,
                    lower,
                    upper,
                    need_lower,
                    need_upper,
                    bigMargs,
                    block,
                    constraint,
                    from_args=True,
                )
                if not need_lower and not need_upper:
                    return lower, upper

        # last check for value for None!
        if None in bigMargs:
            m = bigMargs[None]
            lower, upper, need_lower, need_upper = self._process_M_value(
                m,
                lower,
                upper,
                need_lower,
                need_upper,
                bigMargs,
                None,
                constraint,
                from_args=True,
            )
            if not need_lower and not need_upper:
                return lower, upper

        return lower, upper

    def _estimate_M(self, expr, constraint):
        expr_lb, expr_ub = self._expr_bound_visitor.walk_expression(expr)
        if expr_lb == -interval.inf or expr_ub == interval.inf:
            raise GDP_Error(
                "Cannot estimate M for unbounded "
                "expressions.\n\t(found while processing "
                "constraint '%s'). Please specify a value of M "
                "or ensure all variables that appear in the "
                "constraint are bounded." % constraint.name
            )
        else:
            M = (expr_lb, expr_ub)
        return tuple(M)

    def _add_constraint_expressions(
        self,
        c,
        i,
        M,
        indicator_var,
        newConstraint,
        constraint_map,
        indicator_expression=None,
    ):
        # Since we are both combining components from multiple blocks and using
        # local names, we need to make sure that the first index for
        # transformedConstraints is guaranteed to be unique. We just grab the
        # current length of the list here since that will be monotonically
        # increasing and hence unique. We'll append it to the
        # slightly-more-human-readable constraint name for something familiar
        # but unique. (Note that we really could do this outside of the loop
        # over the constraint indices, but I don't think it matters a lot.)
        unique = len(newConstraint)
        name = c.local_name + "_%s" % unique
        if indicator_expression is None:
            indicator_expression = 1 - indicator_var

        if c.lower is not None:
            if M[0] is None:
                raise GDP_Error(
                    "Cannot relax disjunctive constraint '%s' "
                    "because M is not defined." % name
                )
            M_expr = M[0] * indicator_expression
            newConstraint.add((name, i, 'lb'), c.lower <= c.body - M_expr)
            constraint_map.transformed_constraints[c].append(
                newConstraint[name, i, 'lb']
            )
            constraint_map.src_constraint[newConstraint[name, i, 'lb']] = c
        if c.upper is not None:
            if M[1] is None:
                raise GDP_Error(
                    "Cannot relax disjunctive constraint '%s' "
                    "because M is not defined." % name
                )
            M_expr = M[1] * indicator_expression
            newConstraint.add((name, i, 'ub'), c.body - M_expr <= c.upper)
            constraint_map.transformed_constraints[c].append(
                newConstraint[name, i, 'ub']
            )
            constraint_map.src_constraint[newConstraint[name, i, 'ub']] = c



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.common.collections import ComponentSet, ComponentMap
from pyomo.common.timing import HierarchicalTimer
from pyomo.common.dependencies import attempt_import, numpy as np
from pyomo.core.base.objective import Objective
from pyomo.core.base.suffix import Suffix
from pyomo.core.expr.visitor import identify_variables
from pyomo.util.calc_var_value import calculate_variable_from_constraint
from pyomo.util.subsystems import (
    TemporarySubsystemManager,
    create_subsystem_block,
    generate_subsystem_blocks,
)

# Use attempt_import here due to unguarded NumPy import in these files
pyomo_nlp = attempt_import('pyomo.contrib.pynumero.interfaces.pyomo_nlp')[0]
nlp_proj = attempt_import('pyomo.contrib.pynumero.interfaces.nlp_projections')[0]
from pyomo.contrib.pynumero.algorithms.solvers.cyipopt_solver import CyIpoptSolver
from pyomo.contrib.pynumero.interfaces.cyipopt_interface import CyIpoptNLP
from pyomo.contrib.pynumero.algorithms.solvers.scipy_solvers import (
    FsolveNlpSolver,
    NewtonNlpSolver,
    SecantNewtonNlpSolver,
)
from pyomo.contrib.incidence_analysis import IncidenceGraphInterface
from pyomo.contrib.incidence_analysis.scc_solver import (
    generate_strongly_connected_components,
)


class NlpSolverBase:
    """A base class that solves an NLP object

    Subclasses should implement this interface for compatibility with
    ImplicitFunctionSolver objects.

    """

    def __init__(self, nlp, options=None, timer=None):
        raise NotImplementedError(
            "%s has not implemented the __init__ method" % type(self)
        )

    def solve(self, **kwds):
        raise NotImplementedError(
            "%s has not implemented the solve method" % type(self)
        )


class CyIpoptSolverWrapper(NlpSolverBase):
    """A wrapper for CyIpoptNLP and CyIpoptSolver that implements the
    NlpSolverBase API

    """

    def __init__(self, nlp, options=None, timer=None):
        self._cyipopt_nlp = CyIpoptNLP(nlp)
        self._cyipopt_solver = CyIpoptSolver(self._cyipopt_nlp, options=options)

    def solve(self, **kwds):
        return self._cyipopt_solver.solve(**kwds)


class ScipySolverWrapper(NlpSolverBase):
    """A wrapper for SciPy NLP solvers that implements the NlpSolverBase API

    This solver uses scipy.optimize.fsolve for "vector-valued" NLPs (with more
    than one primal variable and equality constraint) and the Secant-Newton
    hybrid for "scalar-valued" NLPs.

    """

    def __init__(self, nlp, timer=None, options=None):
        if options is None:
            options = {}
        for key in options:
            if (
                key not in SecantNewtonNlpSolver.OPTIONS
                and key not in FsolveNlpSolver.OPTIONS
            ):
                raise ValueError(
                    "Option %s is invalid for both SecantNewtonNlpSolver and"
                    " FsolveNlpSolver" % key
                )
        # Note that options currently contain the options for both solvers.
        # There is currently no way to specify, e.g., different tolerances
        # for the two solvers. This can be updated if there is demand for it.
        newton_options = {
            key: value
            for key, value in options.items()
            if key in SecantNewtonNlpSolver.OPTIONS
        }
        fsolve_options = {
            key: value
            for key, value in options.items()
            if key in FsolveNlpSolver.OPTIONS
        }
        if nlp.n_primals() == 1:
            solver = SecantNewtonNlpSolver(nlp, timer=timer, options=newton_options)
        else:
            solver = FsolveNlpSolver(nlp, timer=timer, options=fsolve_options)
        self._nlp = nlp
        self._options = options
        self._solver = solver

    def solve(self, x0=None):
        res = self._solver.solve(x0=x0)
        return res


class PyomoImplicitFunctionBase:
    """A base class defining an API for implicit functions defined using
    Pyomo components. In particular, this is the API required by
    ExternalPyomoModel.

    Implicit functions are defined by two lists of Pyomo VarData and
    one list of Pyomo ConstraintData. The first list of VarData corresponds
    to "variables" defining the outputs of the implicit function.
    The list of ConstraintData are equality constraints that are converged
    to evaluate the implicit function. The second list of VarData are
    variables to be treated as "parameters" or inputs to the implicit
    function.

    """

    def __init__(self, variables, constraints, parameters):
        """
        Arguments
        ---------
        variables: List of VarData
            Variables to be treated as outputs of the implicit function
        constraints: List of ConstraintData
            Constraints that are converged to evaluate the implicit function
        parameters: List of VarData
            Variables to be treated as inputs to the implicit function

        """
        self._variables = variables
        self._constraints = constraints
        self._parameters = parameters
        self._block_variables = variables + parameters
        self._block = create_subsystem_block(constraints, self._block_variables)

    def get_variables(self):
        return self._variables

    def get_constraints(self):
        return self._constraints

    def get_parameters(self):
        return self._parameters

    def get_block(self):
        return self._block

    def set_parameters(self, values):
        """Sets the parameters of the system that defines the implicit
        function.

        This method does not necessarily need to update values of the Pyomo
        variables, as long as the next evaluation of this implicit function
        is consistent with these inputs.

        Arguments
        ---------
        values: NumPy array
            Array of values to set for the "parameter variables" in the order
            they were specified in the constructor

        """
        raise NotImplementedError()

    def evaluate_outputs(self):
        """Returns the values of the variables that are treated as outputs
        of the implicit function

        The returned values do not necessarily need to be the values stored
        in the Pyomo variables, as long as they are consistent with the
        latest parameters that have been set.

        Returns
        -------
        NumPy array
            Array with values corresponding to the "output variables" in
            the order they were specified in the constructor

        """
        raise NotImplementedError()

    def update_pyomo_model(self):
        """Sets values of "parameter variables" and "output variables"
        to the most recent values set or computed in this implicit function

        """
        raise NotImplementedError()


class ImplicitFunctionSolver(PyomoImplicitFunctionBase):
    """A basic implicit function solver that uses a ProjectedNLP to solve
    the parameterized system without repeated file writes when parameters
    are updated

    """

    def __init__(
        self,
        variables,
        constraints,
        parameters,
        solver_class=None,
        solver_options=None,
        timer=None,
    ):
        if timer is None:
            timer = HierarchicalTimer()
        self._timer = timer
        if solver_options is None:
            solver_options = {}

        self._timer.start("__init__")

        super().__init__(variables, constraints, parameters)
        block = self.get_block()

        # PyomoNLP requires an objective
        block._obj = Objective(expr=0.0)
        # CyIpoptSolver requires a non-empty scaling factor
        block.scaling_factor = Suffix(direction=Suffix.EXPORT)
        block.scaling_factor[block._obj] = 1.0

        self._timer.start("PyomoNLP")
        self._nlp = pyomo_nlp.PyomoNLP(block)
        self._timer.stop("PyomoNLP")
        primals_ordering = [var.name for var in variables]
        self._proj_nlp = nlp_proj.ProjectedExtendedNLP(self._nlp, primals_ordering)

        self._timer.start("NlpSolver")
        if solver_class is None:
            self._solver = ScipySolverWrapper(
                self._proj_nlp, options=solver_options, timer=timer
            )
        else:
            self._solver = solver_class(
                self._proj_nlp, options=solver_options, timer=timer
            )
        self._timer.stop("NlpSolver")

        vars_in_cons = []
        _seen = set()
        for con in constraints:
            for var in identify_variables(con.expr, include_fixed=False):
                if id(var) not in _seen:
                    _seen.add(id(var))
                    vars_in_cons.append(var)
        self._active_var_set = ComponentSet(vars_in_cons)

        # It is possible (and fairly common) for variables specified as
        # parameters to not appear in any of the specified constraints.
        # We will fail if we try to get their coordinates in the NLP.
        #
        # Technically, this could happen for the variables as well. However,
        # this would guarantee that the Jacobian is singular. I will worry
        # about this when I encounter such a case.
        self._active_param_mask = np.array(
            [(p in self._active_var_set) for p in parameters]
        )
        self._active_parameters = [
            p for i, p in enumerate(parameters) if self._active_param_mask[i]
        ]
        if any((var not in self._active_var_set) for var in variables):
            raise RuntimeError(
                "Invalid model. All variables must appear in specified constraints."
            )

        # These are coordinates in the original NLP
        self._variable_coords = self._nlp.get_primal_indices(variables)
        self._active_parameter_coords = self._nlp.get_primal_indices(
            self._active_parameters
        )

        # NOTE: With this array, we are storing the same data in two locations.
        # Once here, and once in the NLP. We do this because parameters do not
        # *need* to exist in the NLP. However, we still need to be able to
        # update the "parameter variables" in the Pyomo model. If we only stored
        # the parameters in the NLP, we would lose the values for parameters
        # that don't appear in the active constraints.
        self._parameter_values = np.array([var.value for var in parameters])

        self._timer.start("__init__")

    def set_parameters(self, values, **kwds):
        self._timer.start("set_parameters")
        # I am not 100% sure the values will always be an array (as opposed to
        # list), so explicitly convert here.
        values = np.array(values)
        self._parameter_values = values
        values = np.compress(self._active_param_mask, values)
        primals = self._nlp.get_primals()
        # Will it cause a problem that _active_parameter_coords is a list
        # rather than array?
        primals[self._active_parameter_coords] = values
        self._nlp.set_primals(primals)
        self._timer.start("solve")
        results = self._solver.solve(**kwds)
        self._timer.stop("solve")
        self._timer.stop("set_parameters")
        return results

    def evaluate_outputs(self):
        primals = self._nlp.get_primals()
        outputs = primals[self._variable_coords]
        return outputs

    def update_pyomo_model(self):
        primals = self._nlp.get_primals()
        for i, var in enumerate(self.get_variables()):
            var.set_value(primals[self._variable_coords[i]], skip_validation=True)
        for var, value in zip(self._parameters, self._parameter_values):
            var.set_value(value, skip_validation=True)


class DecomposedImplicitFunctionBase(PyomoImplicitFunctionBase):
    """A base class for an implicit function that applies a partition
    to its variables and constraints and converges the system by solving
    subsets sequentially

    Subclasses should implement the partition_system method, which
    determines how variables and constraints are partitioned into subsets.

    """

    def __init__(
        self,
        variables,
        constraints,
        parameters,
        solver_class=None,
        solver_options=None,
        timer=None,
        use_calc_var=True,
    ):
        if timer is None:
            timer = HierarchicalTimer()
        self._timer = timer
        self._timer.start("__init__")
        if solver_class is None:
            solver_class = ScipySolverWrapper
        self._solver_class = solver_class
        if solver_options is None:
            solver_options = {}
        self._solver_options = solver_options
        self._calc_var_cutoff = 1 if use_calc_var else 0
        # NOTE: This super call is only necessary so the get_* methods work
        super().__init__(variables, constraints, parameters)

        subsystem_list = [
            # Switch order in list for compatibility with generate_subsystem_blocks
            (cons, vars)
            for vars, cons in self.partition_system(variables, constraints)
        ]

        var_param_set = ComponentSet(variables + parameters)
        # We will treat variables that are neither variables nor parameters
        # as "constants". These are usually things like area, volume, or some
        # other global parameter that is treated as a variable and "fixed" with
        # an equality constraint.
        constants = []
        constant_set = ComponentSet()
        for con in constraints:
            for var in identify_variables(con.expr, include_fixed=False):
                if var not in constant_set and var not in var_param_set:
                    # If this is a newly encountered variable that is neither
                    # a var nor param, treat it as a "constant"
                    constant_set.add(var)
                    constants.append(var)

        with TemporarySubsystemManager(to_fix=constants):
            # Temporarily fix "constant" variables so (a) they don't show
            # up in the local inputs of the subsystem blocks and (b) so
            # they don't appear as additional columns in the NLPs and
            # ProjectedNLPs.

            self._subsystem_list = list(generate_subsystem_blocks(subsystem_list))
            # These are subsystems that need an external solver, rather than
            # calculate_variable_from_constraint. _calc_var_cutoff should be either
            # 0 or 1.
            self._solver_subsystem_list = [
                (block, inputs)
                for block, inputs in self._subsystem_list
                if len(block.vars) > self._calc_var_cutoff
            ]

            # Need a dummy objective to create an NLP
            for block, inputs in self._solver_subsystem_list:
                block._obj = Objective(expr=0.0)
                # I need scaling_factor so Pyomo NLPs I create from these blocks
                # don't break when ProjectedNLP calls get_primals_scaling
                block.scaling_factor = Suffix(direction=Suffix.EXPORT)
                # HACK: scaling_factor just needs to be nonempty
                block.scaling_factor[block._obj] = 1.0

            # Original PyomoNLP for each subset in the partition
            # Since we are creating these NLPs with "constants" fixed, these
            # variables will not show up in the NLPs
            self._timer.start("PyomoNLP")
            self._solver_subsystem_nlps = [
                pyomo_nlp.PyomoNLP(block)
                for block, inputs in self._solver_subsystem_list
            ]
            self._timer.stop("PyomoNLP")

        # "Output variable" names are required to construct ProjectedNLPs.
        # Ideally, we can eventually replace these with variable indices.
        self._solver_subsystem_var_names = [
            [var.name for var in block.vars.values()]
            for block, inputs in self._solver_subsystem_list
        ]
        self._solver_proj_nlps = [
            nlp_proj.ProjectedExtendedNLP(nlp, names)
            for nlp, names in zip(
                self._solver_subsystem_nlps, self._solver_subsystem_var_names
            )
        ]

        # We will solve the ProjectedNLPs rather than the original NLPs
        self._timer.start("NlpSolver")
        self._nlp_solvers = [
            self._solver_class(nlp, timer=self._timer, options=self._solver_options)
            for nlp in self._solver_proj_nlps
        ]
        self._timer.stop("NlpSolver")
        self._solver_subsystem_input_coords = [
            # Coordinates in the NLP, not ProjectedNLP
            nlp.get_primal_indices(inputs)
            for nlp, (subsystem, inputs) in zip(
                self._solver_subsystem_nlps, self._solver_subsystem_list
            )
        ]

        self._n_variables = len(variables)
        self._n_constraints = len(constraints)
        self._n_parameters = len(parameters)

        # This is a global (wrt individual subsystems) array that stores
        # the current values of variables and parameters. This is useful
        # for updating values in between subsystem solves.
        #
        # NOTE: This could also be implemented as a tuple of
        # (subsystem_coord, primal_coord), which would eliminate the need to
        # store data in two locations. The current implementation is easier,
        # however.
        self._global_values = np.array([var.value for var in variables + parameters])
        self._global_indices = ComponentMap(
            (var, i) for i, var in enumerate(variables + parameters)
        )
        # Cache the global array-coordinates of each subset of "input"
        # variables. These are used for updating before each solve.
        self._local_input_global_coords = [
            # If I do not fix "constants" above, I get errors here
            # that only show up in the CLC models.
            # TODO: Add a test that covers this edge case.
            np.array([self._global_indices[var] for var in inputs], dtype=int)
            for (_, inputs) in self._solver_subsystem_list
        ]

        # Cache the global array-coordinates of each subset of "output"
        # variables. These are used for updating after each solve.
        self._output_coords = [
            np.array(
                [self._global_indices[var] for var in block.vars.values()], dtype=int
            )
            for (block, _) in self._solver_subsystem_list
        ]

        self._timer.stop("__init__")

    def n_subsystems(self):
        """Returns the number of subsystems in the partition of variables
        and equations used to converge the system defining the implicit
        function

        """
        return len(self._subsystem_list)

    def partition_system(self, variables, constraints):
        """Partitions the systems of equations defined by the provided
        variables and constraints

        Each subset of the partition should have an equal number of variables
        and equations. These subsets, or "subsystems", will be solved
        sequentially in the order provided by this method instead of solving
        the entire system simultaneously. Subclasses should implement this
        method to define the partition that their implicit function solver
        will use. Partitions are defined as a list of tuples of lists.
        Each tuple has two entries, the first a list of variables, and the
        second a list of constraints. These inner lists should have the
        same number of entries.

        Arguments
        ---------
        variables: list
            List of VarData in the system to be partitioned
        constraints: list
            List of ConstraintData (equality constraints) defining the
            equations of the system to be partitioned

        Returns
        -------
        List of tuples
            List of tuples describing the ordered partition. Each tuple
            contains equal-length subsets of variables and constraints.

        """
        # Subclasses should implement this method, which returns an ordered
        # partition (two lists-of-lists) of variables and constraints.
        raise NotImplementedError(
            "%s has not implemented the partition_system method" % type(self)
        )

    def set_parameters(self, values):
        self._timer.start("set_parameters")
        values = np.array(values)
        #
        # Set parameter values
        #
        # NOTE: Here I rely on the fact that the "global array" is in the
        # order (variables, parameters)
        self._global_values[self._n_variables :] = values

        #
        # Solve subsystems one-by-one
        #
        # The basic procedure is: update local information from the global
        # array, solve the subsystem, then update the global array with
        # new values.
        solver_subsystem_idx = 0
        for block, inputs in self._subsystem_list:
            if len(block.vars) <= self._calc_var_cutoff:
                # Update model values from global array.
                for var in inputs:
                    idx = self._global_indices[var]
                    var.set_value(self._global_values[idx], skip_validation=True)
                # Solve using calculate_variable_from_constraint
                var = block.vars[0]
                con = block.cons[0]
                self._timer.start("solve")
                self._timer.start("calc_var")
                calculate_variable_from_constraint(var, con)
                self._timer.stop("calc_var")
                self._timer.stop("solve")
                # Update global array with values from solve
                self._global_values[self._global_indices[var]] = var.value
            else:
                # Transfer variable values into the projected NLP, solve,
                # and extract values.

                i = solver_subsystem_idx
                nlp = self._solver_subsystem_nlps[i]
                proj_nlp = self._solver_proj_nlps[i]
                input_coords = self._solver_subsystem_input_coords[i]
                input_global_coords = self._local_input_global_coords[i]
                output_global_coords = self._output_coords[i]

                nlp_solver = self._nlp_solvers[solver_subsystem_idx]

                # Get primals, load potentially new input values into primals,
                # then load primals into NLP
                primals = nlp.get_primals()
                primals[input_coords] = self._global_values[input_global_coords]

                # Set primals in the original NLP. This is necessary so the
                # parameters get updated.
                nlp.set_primals(primals)

                # Get initial guess in the space of variables we solve for
                x0 = proj_nlp.get_primals()
                self._timer.start("solve")
                self._timer.start("solve_nlp")
                nlp_solver.solve(x0=x0)
                self._timer.stop("solve_nlp")
                self._timer.stop("solve")

                # Set values in global array. Here we rely on the fact that
                # the projected NLP's primals are in the order that variables
                # were initially specified.
                self._global_values[output_global_coords] = proj_nlp.get_primals()

                solver_subsystem_idx += 1

        self._timer.stop("set_parameters")

    def evaluate_outputs(self):
        return self._global_values[: self._n_variables]

    def update_pyomo_model(self):
        # NOTE: Here we rely on the fact that global_values is in the
        # order (variables, parameters)
        for i, var in enumerate(self.get_variables() + self.get_parameters()):
            var.set_value(self._global_values[i], skip_validation=True)


class SccImplicitFunctionSolver(DecomposedImplicitFunctionBase):
    def partition_system(self, variables, constraints):
        self._timer.start("partition")
        igraph = IncidenceGraphInterface()
        var_blocks, con_blocks = igraph.block_triangularize(variables, constraints)
        self._timer.stop("partition")
        return zip(var_blocks, con_blocks)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

"""
Script to generate the installer for pyomo.
"""

import os
import platform
import sys
from pathlib import Path
from setuptools import setup, find_packages, Command

try:
    # This works beginning in setuptools 40.7.0 (27 Jan 2019)
    from setuptools import DistutilsOptionError
except ImportError:
    # Needed for setuptools prior to 40.7.0
    from distutils.errors import DistutilsOptionError


def read(*rnames):
    with open(os.path.join(os.path.dirname(__file__), *rnames)) as README:
        # Strip all leading badges up to, but not including the COIN-OR
        # badge so that they do not appear in the PyPI description
        while True:
            line = README.readline()
            if 'COIN-OR' in line:
                break
            if line.strip() and '[![' not in line:
                break
        return line + README.read()


def import_pyomo_module(*path):
    _module_globals = dict(globals())
    _module_globals['__name__'] = None
    _source = os.path.join(os.path.dirname(__file__), *path)
    with open(_source) as _FILE:
        exec(_FILE.read(), _module_globals)
    return _module_globals


def get_version():
    # Source pyomo/version/info.py to get the version number
    return import_pyomo_module('pyomo', 'version', 'info.py')['__version__']


def check_config_arg(name):
    if name in sys.argv:
        sys.argv.remove(name)
        return True
    if name in os.getenv('PYOMO_SETUP_ARGS', '').split():
        return True
    return False


CYTHON_REQUIRED = "required"
if not any(
    arg.startswith(cmd)
    for cmd in ('build', 'install', 'bdist', 'wheel')
    for arg in sys.argv
):
    using_cython = False
elif sys.version_info[:2] < (3, 11):
    using_cython = "automatic"
else:
    using_cython = False
if check_config_arg('--with-cython'):
    using_cython = CYTHON_REQUIRED
if check_config_arg('--without-cython'):
    using_cython = False

ext_modules = []
if using_cython:
    try:
        if platform.python_implementation() != "CPython":
            # break out of this try-except (disable Cython)
            raise RuntimeError("Cython is only supported under CPython")
        from Cython.Build import cythonize

        #
        # Note: The Cython developers recommend that you distribute C source
        # files to users.  But this is fine for evaluating the utility of Cython
        #
        import shutil

        files = [
            "pyomo/core/expr/numvalue.pyx",
            "pyomo/core/expr/numeric_expr.pyx",
            "pyomo/core/expr/logical_expr.pyx",
            # "pyomo/core/expr/visitor.pyx",
            "pyomo/core/util.pyx",
            "pyomo/repn/standard_repn.pyx",
            "pyomo/repn/plugins/cpxlp.pyx",
            "pyomo/repn/plugins/gams_writer.pyx",
            "pyomo/repn/plugins/baron_writer.pyx",
            "pyomo/repn/plugins/ampl/ampl_.pyx",
        ]
        for f in files:
            shutil.copyfile(f[:-1], f)
        ext_modules = cythonize(files, compiler_directives={"language_level": 3})
    except:
        if using_cython == CYTHON_REQUIRED:
            print("""
ERROR: Cython was explicitly requested with --with-cython, but cythonization
       of core Pyomo modules failed.
""")
            raise
        using_cython = False

if check_config_arg('--with-distributable-extensions'):
    #
    # Import the APPSI extension builder
    # NOTE: There is inconsistent behavior in Windows for APPSI.
    # As a result, we will NOT include these extensions in Windows.
    if not sys.platform.startswith('win'):
        appsi_extension = import_pyomo_module('pyomo', 'contrib', 'appsi', 'build.py')[
            'get_appsi_extension'
        ](
            in_setup=True,
            appsi_root=os.path.join(
                os.path.dirname(__file__), 'pyomo', 'contrib', 'appsi'
            ),
        )
        ext_modules.append(appsi_extension)


class DependenciesCommand(Command):
    """Custom setuptools command

    This will output the list of dependencies, including any optional
    dependencies for 'extras_require` targets.  This is needed so that
    we can (relatively) easily extract what `pip install '.[optional]'`
    would have done so that we can pass it on to a 'conda install'
    command when setting up Pyomo testing in a conda environment
    (because conda for all intents does not acknowledge
    `extras_require`).

    """

    description = "list the dependencies for this package"
    user_options = [('extras=', None, 'extra targets to include')]

    def initialize_options(self):
        self.extras = None

    def finalize_options(self):
        if self.extras is not None:
            self.extras = [e for e in (_.strip() for _ in self.extras.split(',')) if e]
            for e in self.extras:
                if e not in setup_kwargs['extras_require']:
                    raise DistutilsOptionError(
                        "extras can only include {%s}"
                        % (', '.join(setup_kwargs['extras_require']))
                    )

    def run(self):
        deps = list(self._print_deps(setup_kwargs['install_requires']))
        if self.extras is not None:
            for e in self.extras:
                deps.extend(self._print_deps(setup_kwargs['extras_require'][e]))
        print(' '.join(deps))

    def _print_deps(self, deplist):
        class version_cmp:
            ver = tuple(map(int, platform.python_version_tuple()[:2]))

            def __lt__(self, other):
                return self.ver < tuple(map(int, other.split('.')))

            def __le__(self, other):
                return self.ver <= tuple(map(int, other.split('.')))

            def __gt__(self, other):
                return not self.__le__(other)

            def __ge__(self, other):
                return not self.__lt__(other)

            def __eq__(self, other):
                return self.ver == tuple(map(int, other.split('.')))

            def __ne__(self, other):
                return not self.__eq__(other)

        implementation_name = sys.implementation.name
        platform_system = platform.system()
        python_version = version_cmp()
        for entry in deplist:
            dep, _, condition = (_.strip() for _ in entry.partition(';'))
            if condition and not eval(condition):
                continue
            yield dep


setup_kwargs = dict(
    cmdclass={'dependencies': DependenciesCommand},
    version=get_version(),
    install_requires=[],
    extras_require={
        # There are certain tests that also require pytest-qt, but because those
        # tests are so environment/machine specific, we are leaving these out of
        # the dependencies.
        'tests': [
            'coverage',
            'parameterized',
            'pybind11',
            # 9.0.0 breaks skipping individual tests; see
            # https://github.com/pytest-dev/pytest/issues/13895
            'pytest!=9.0.0',
            'pytest-parallel',
        ],
        'docs': [
            # Sphinx 9.0-9.1.0 fails to correctly generate type references.
            'Sphinx>4,!=8.2.0,!=9.0.*,!=9.1.0',
            'sphinx-copybutton',
            'sphinx_rtd_theme>0.5',
            'sphinxcontrib-jsmath',
            'sphinxcontrib-napoleon',
            'numpy',  # Needed by autodoc for pynumero
            'scipy',  # Needed by autodoc for pynumero
        ],
        'optional': [
            'dill',  # No direct use, but improves lambda pickle
            'ipython',  # contrib.viewer
            'linear-tree; python_version<"3.14"',  # contrib.piecewise
            # FIXME: This is a temporary pin that should be removed
            # when the linear-tree dependency is replaced
            'scikit-learn<1.7.0; implementation_name!="pypy" and python_version<"3.14"',
            'scikit-learn; implementation_name!="pypy" and python_version>="3.14"',
            # Note: matplotlib 3.6.1 has bug #24127, which breaks
            # seaborn's histplot (triggering parmest failures)
            # Note: minimum version from community_detection use of
            # matplotlib.pyplot.get_cmap()
            'matplotlib>=3.6.0,!=3.6.1',
            'networkx',  # network, incidence_analysis, community_detection
            'numpy',
            'openpyxl',  # dataportals
            'packaging',  # for checking other dependency versions
            #'pathos',   # requested for #963, but PR currently closed
            # pint causes a segfault only for one test (test_bad_units) on pypy
            'pint; implementation_name!="pypy"',  # units
            'plotly',  # incidence_analysis
            'python-louvain',  # community_detection
            'pyyaml',  # core
            # qtconsole also requires a supported Qt version (PyQt5 or PySide6).
            # Because those are environment specific, we have left that out here.
            'qtconsole',  # contrib.viewer
            'scipy',
            'sympy',  # differentiation
            'xlrd',  # dataportals
            'z3-solver',  # community_detection
            #
            # subprocess output is merged more reliably if
            # 'PeekNamedPipe' is available from pywin32
            'pywin32; platform_system=="Windows"',
            #
            # The following optional dependencies are difficult to
            # install on PyPy (binary wheels are not available), so we
            # will only "require" them on other (CPython) platforms:
            'casadi; implementation_name!="pypy"',  # dae
            'numdifftools; implementation_name!="pypy"',  # pynumero
            'pandas; implementation_name!="pypy"',
            'seaborn; implementation_name!="pypy"',  # parmest.graphics
        ],
    },
    packages=find_packages(exclude=("scripts",)),
    package_data={
        "pyomo.contrib.ampl_function_demo": ["src/*"],
        "pyomo.contrib.appsi.cmodel": ["src/*"],
        "pyomo.contrib.cspline_external": ["src/*"],
        "pyomo.contrib.aslfunctions": ["src/*"],
        "pyomo.contrib.mcpp": ["*.cpp"],
        "pyomo.contrib.pynumero": ['src/*', 'src/tests/*'],
        "pyomo.contrib.viewer": ["*.ui"],
        "pyomo.contrib.simplification.ginac": ["src/*.cpp", "src/*.hpp"],
    },
    ext_modules=ext_modules,
)


try:
    # setuptools.build_meta (>=68) forbids absolute paths in the `sources=` list.
    # This resets the extensions (only for those items that are absolute paths)
    # to use relative paths
    ROOT = Path(__file__).parent.resolve()
    for ext in ext_modules:
        rel_sources = []
        for src in ext.sources:
            p = Path(src)
            if p.is_absolute():
                p = p.relative_to(ROOT)
            rel_sources.append(p.as_posix())
        ext.sources[:] = rel_sources
    setup(**setup_kwargs)
except SystemExit as e_info:
    # Cython can generate a SystemExit exception on Windows if the
    # environment is missing / has an incorrect Microsoft compiler.
    # Since Cython is not strictly required, we will disable Cython and
    # try re-running setup(), but only for this very specific situation.
    if 'Microsoft Visual C++' not in str(e_info):
        raise
    elif using_cython == CYTHON_REQUIRED:
        print("""
ERROR: Cython was explicitly requested with --with-cython, but cythonization
       of core Pyomo modules failed.
""")
        raise
    else:
        print("""
ERROR: setup() failed:
    %s
Re-running setup() without the Cython modules
""" % (str(e_info),))
        setup_kwargs['ext_modules'] = []
        setup(**setup_kwargs)
        print("""
WARNING: Installation completed successfully, but the attempt to cythonize
         core Pyomo modules failed.  Cython provides performance
         optimizations and is not required for any Pyomo functionality.
         Cython returned the following error:
   "%s"
""" % (str(e_info),))
